from laktory.models import Catalog


catalog = Catalog(
    name="dev",
    grants=[
        {"principal": "account users", "privileges": ["USE_CATALOG", "USE_SCHEMA"]}
    ],
    schemas=[
        {
            "name": "engineering",
            "grants": [{"principal": "domain-engineering", "privileges": ["SELECT"]}],
        },
        {
            "name": "sources",
            "volumes": [
                {
                    "name": "landing",
                    "volume_type": "EXTERNAL",
                    "grants": [
                        {
                            "principal": "account users",
                            "privileges": ["READ_VOLUME"],
                        },
                        {
                            "principal": "role-metastore-admins",
                            "privileges": ["WRITE_VOLUME"],
                        },
                    ],
                },
            ],
        },
    ],
)


def test_model():
    assert catalog.name == "dev"
    assert catalog.full_name == "dev"
    assert catalog.full_name == "dev"


if __name__ == "__main__":
    test_model()



from typing import Union
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MetastoreAssignment(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Metastore Assignment

    Attributes
    ----------
    metastore_id:
        Unique identifier of the parent Metastore
    workspace_id:
        id of the workspace for the assignment
    default_catalog_name:
        Default catalog used for this assignment, default to hive_metastore

    Examples
    --------
    ```py
    ```
    """

    default_catalog_name: str = None
    metastore_id: Union[int, str] = None
    workspace_id: Union[int, str] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self):
        return "assignment"

    @property
    def resource_key(self):
        return f"{self.metastore_id}-{self.workspace_id}"

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MetastoreAssignment"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.MetastoreAssignment

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_metastore_assignment"



from pyspark.sql import Row
from pyspark.sql import SparkSession
import os
import pytest

from laktory.models.datasources import EventDataSource
from laktory.models.datasources import TableDataSource
from laktory._testing import StockPriceDataEventHeader
from laktory._testing import EventsManager


# Build and write events
data_dir = os.path.join(os.path.dirname(__file__), "data/")
header = StockPriceDataEventHeader()
manager = EventsManager()
manager.build_events(data_dir)
manager.to_path()

# Spark
spark = SparkSession.builder.appName("UnitTesting").getOrCreate()


def test_event_data_source():
    source = EventDataSource(
        name="stock_price",
        producer={"name": "yahoo_finance"},
    )
    assert (
        source.event_root
        == "/Volumes/dev/sources/landing/events/yahoo_finance/stock_price/"
    )
    assert not source.is_cdc


def test_event_data_source_read():
    source = EventDataSource(
        name="stock_price",
        producer={"name": "yahoo-finance"},
        events_root=data_dir,
        read_as_stream=False,
    )
    df = source.read(spark).toPandas()
    assert len(df) == 80
    print(list(df.columns))
    assert list(df.columns) == [
        "data",
        "description",
        "event_root_",
        "name",
        "producer",
    ]
    df["data"] = df["data"].apply(Row.asDict)
    df["symbol"] = df["data"].apply(dict.get, args=("symbol",))
    df["created_at"] = df["data"].apply(dict.get, args=("_created_at",))
    df = df.sort_values(["symbol", "created_at"])
    row = df.iloc[0]["data"]
    assert row["symbol"] == "AAPL"
    assert row["close"] == pytest.approx(189.46, abs=0.01)


def test_table_data_souurce():
    source = TableDataSource(
        path="/Volumes/tables/stock_prices/",
    )
    assert source.path == "/Volumes/tables/stock_prices/"
    assert source.from_path
    assert source.path_or_full_name == "/Volumes/tables/stock_prices/"


if __name__ == "__main__":
    test_event_data_source()
    test_event_data_source_read()
    test_table_data_souurce()



from laktory.models import User
from laktory.models import Group

user = User(
    user_name="data.engineer@gmail.moc",
    group_ids=[
        "${resources.group-role-engineers.id}",
    ],
    roles=["store-admin"],
)

group = Group(
    display_name="role-engineers",
)


def test_user_group():
    assert user.roles == ["store-admin"]
    assert group.display_name == "role-engineers"


if __name__ == "__main__":
    test_user_group()



import os
from typing import Any
from typing import Literal
from typing import Union
from pydantic import model_validator
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.accesscontrol import AccessControl
from laktory.models.databricks.permissions import Permissions


class SqlQuery(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks SQL Query

    Attributes
    ----------
    access_controls:
        SQL Query access controls
    comment:
        General description that conveys additional information about this query such as usage notes.
    data_source_id:
        Data source ID of the SQL warehouse that will be used to run the query.
    name:
        The title of this query that appears in list views, widget headings, and on the query page.
    parent:
        Id of the workspace folder storing the query
    query:
        The text of the query to be run.
    run_as_role:
        Run as role.
    tags:
        List of tags

    Examples
    --------
    ```py
    from laktory import models

    q = models.SqlQuery(
        name="create-view",
        query="CREATE VIEW google_stock_prices AS SELECT * FROM stock_prices WHERE symbol = 'GOOGL'",
        data_source_id="09z739ce103q9374",
        parent="folders/2479128258235163",
        access_controls=[
            {
                "group_name": "role-engineers",
                "permission_level": "CAN_RUN",
            }
        ],
    )
    ```
    """

    access_controls: list[AccessControl] = []
    comment: str = None
    data_source_id: str
    name: str
    parent: str
    query: str
    run_as_role: Literal["viewer", "owner"] = None
    tags: list[str] = []

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    sql_query_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SqlQuery"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.SqlQuery

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "warehouse_id"]

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"comment": "description"}

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_sql_query"

    @property
    def terraform_renames(self) -> dict[str, str]:
        return self.pulumi_renames

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import pulumi
import os

from laktory.models import BaseModel
from laktory.models import Table
from laktory.models import Schema
from laktory.models import Job
from laktory import settings
from pulumi_random import RandomString

dirpath = os.path.dirname(__file__)

env = RandomString("env", length=3, upper=False, numeric=False, special=False)
schema_name = RandomString(
    "schema", length=5, upper=False, numeric=False, special=False
)


schema = Schema(
    name="${vars.env}.${vars.schema_name}",
    catalog_name="${vars.env}",
    tables=[
        Table(
            name="AAPL",
            columns=[
                {
                    "name": "open",
                    "type": "double",
                    "spark_func_kwargs": {
                        "window_length": 2
                    },  # window_length should not be converted to camel
                },
                {
                    "name": "${resources.close.id}",
                    "type": "double",
                },
            ],
        ),
        Table(
            name="GOOGL",
            columns=[
                {
                    "name": "${vars.dynamic_column}",
                    "type": "double",
                },
                {
                    "name": "high",
                    "type": "double",
                },
            ],
        ),
    ],
    variables={
        "DYNAMIC_COLUMN": "low",
        "env": env.id,
        "schema_name": schema_name.id,
        # r"\$\{resources\.([\w.]+)\}": r"\1",
        r"\$\{resources\.([\w.]+)\}": r"${ \1 }",
    },
)


def test_read_yaml():
    class OHLC(BaseModel):
        open: float = None
        high: float = None
        low: float = None
        close: float = None

    class Price(BaseModel):
        timestamp: float
        ohlc: OHLC

    class StockPrices(BaseModel):
        symbol: str
        prices: list[Price]

    with open(os.path.join(dirpath, "stockprices0.yaml"), "r") as fp:
        stockprices = StockPrices.model_validate_yaml(fp)

    assert stockprices.model_dump() == {
        "symbol": "AAPL",
        "prices": [
            {
                "timestamp": 1.0,
                "ohlc": {"open": 0.0, "high": None, "low": None, "close": 1.0},
            },
            {
                "timestamp": 2.0,
                "ohlc": {"open": 1.0, "high": None, "low": None, "close": 2.0},
            },
            {
                "timestamp": 3.0,
                "ohlc": {"open": None, "high": None, "low": 3.0, "close": 4.0},
            },
        ],
    }


def test_camelize():
    settings.camel_serialization = True
    dump = schema.model_dump()
    print(dump)
    assert dump == {
        "comment": None,
        "grants": None,
        "name": "${vars.env}.${vars.schema_name}",
        "tables": [
            {
                "builder": {
                    "aggregation": None,
                    "filter": None,
                    "joins": [],
                    "layer": None,
                    "selects": None,
                    "template": None,
                    "asDltView": False,
                    "dropColumns": [],
                    "dropDuplicates": None,
                    "dropSourceColumns": None,
                    "eventSource": None,
                    "joinsPostAggregation": [],
                    "pipelineName": None,
                    "tableSource": None,
                    "windowFilter": None,
                },
                "columns": [
                    {
                        "comment": None,
                        "name": "open",
                        "pii": None,
                        "type": "double",
                        "unit": None,
                        "catalogName": None,
                        "schemaName": None,
                        "sparkFuncArgs": [],
                        "sparkFuncKwargs": {
                            "window_length": {
                                "value": 2,
                                "isColumn": False,
                                "toLit": True,
                                "toExpr": False,
                            }
                        },
                        "sparkFuncName": None,
                        "sqlExpression": None,
                        "tableName": "AAPL",
                    },
                    {
                        "comment": None,
                        "name": "${resources.close.id}",
                        "pii": None,
                        "type": "double",
                        "unit": None,
                        "catalogName": None,
                        "schemaName": None,
                        "sparkFuncArgs": [],
                        "sparkFuncKwargs": {},
                        "sparkFuncName": None,
                        "sqlExpression": None,
                        "tableName": "AAPL",
                    },
                ],
                "comment": None,
                "data": None,
                "expectations": [],
                "grants": None,
                "name": "AAPL",
                "catalogName": "${vars.env}",
                "dataSourceFormat": "DELTA",
                "primaryKey": None,
                "schemaName": "${vars.env}.${vars.schema_name}",
                "tableType": "MANAGED",
                "timestampKey": None,
                "viewDefinition": None,
                "warehouseId": "08b717ce051a0261",
            },
            {
                "builder": {
                    "aggregation": None,
                    "filter": None,
                    "joins": [],
                    "layer": None,
                    "selects": None,
                    "template": None,
                    "asDltView": False,
                    "dropColumns": [],
                    "dropDuplicates": None,
                    "dropSourceColumns": None,
                    "eventSource": None,
                    "joinsPostAggregation": [],
                    "pipelineName": None,
                    "tableSource": None,
                    "windowFilter": None,
                },
                "columns": [
                    {
                        "comment": None,
                        "name": "${vars.dynamic_column}",
                        "pii": None,
                        "type": "double",
                        "unit": None,
                        "catalogName": None,
                        "schemaName": None,
                        "sparkFuncArgs": [],
                        "sparkFuncKwargs": {},
                        "sparkFuncName": None,
                        "sqlExpression": None,
                        "tableName": "GOOGL",
                    },
                    {
                        "comment": None,
                        "name": "high",
                        "pii": None,
                        "type": "double",
                        "unit": None,
                        "catalogName": None,
                        "schemaName": None,
                        "sparkFuncArgs": [],
                        "sparkFuncKwargs": {},
                        "sparkFuncName": None,
                        "sqlExpression": None,
                        "tableName": "GOOGL",
                    },
                ],
                "comment": None,
                "data": None,
                "expectations": [],
                "grants": None,
                "name": "GOOGL",
                "catalogName": "${vars.env}",
                "dataSourceFormat": "DELTA",
                "primaryKey": None,
                "schemaName": "${vars.env}.${vars.schema_name}",
                "tableType": "MANAGED",
                "timestampKey": None,
                "viewDefinition": None,
                "warehouseId": "08b717ce051a0261",
            },
        ],
        "volumes": [],
        "catalogName": "${vars.env}",
        "forceDestroy": True,
    }

    settings.camel_serialization = False


def test_singular():
    job = Job(
        name="my-job",
        clusters=[
            {
                "name": "main",
                "spark_version": "14.0.x-scala2.12",
                "node_type_id": "${vars.node_type_id}",
                "spark_env_vars": {
                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                    "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                },
            }
        ],
    )

    settings.singular_serialization = True
    dump = job.model_dump()
    print(dump)
    assert dump == {
        "continuous": None,
        "control_run_state": None,
        "email_notifications": None,
        "format": None,
        "health": None,
        "max_concurrent_runs": None,
        "max_retries": None,
        "min_retry_interval_millis": None,
        "name": "my-job",
        "notification_settings": None,
        "retry_on_timeout": None,
        "run_as": None,
        "schedule": None,
        "tags": {},
        "timeout_seconds": None,
        "trigger": None,
        "webhook_notifications": None,
        "access_control": [],
        "cluster": [
            {
                "apply_policy_default_values": None,
                "autoscale": None,
                "autotermination_minutes": None,
                "cluster_id": None,
                "custom_tags": None,
                "data_security_mode": "USER_ISOLATION",
                "driver_instance_pool_id": None,
                "driver_node_type_id": None,
                "enable_elastic_disk": None,
                "enable_local_disk_encryption": None,
                "idempotency_token": None,
                "init_scripts": [],
                "instance_pool_id": None,
                "name": "main",
                "node_type_id": "${vars.node_type_id}",
                "num_workers": None,
                "policy_id": None,
                "runtime_engine": None,
                "single_user_name": None,
                "spark_conf": {},
                "spark_env_vars": {
                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                    "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                },
                "spark_version": "14.0.x-scala2.12",
                "ssh_public_keys": [],
            }
        ],
        "parameter": [],
        "task": [],
    }
    settings.singular_serialization = False


def test_inject_vars():
    d0 = schema.model_dump()
    d1 = schema.inject_vars(d0)
    assert d1["tables"][-1]["columns"][0]["name"] == "low"
    assert isinstance(d1["name"], pulumi.Output)
    assert isinstance(d1["catalog_name"], pulumi.Output)
    assert d1["tables"][0]["columns"][1]["name"] == "${ close.id }"


if __name__ == "__main__":
    test_read_yaml()
    test_camelize()
    test_singular()
    test_inject_vars()



from pyspark.sql import SparkSession
from pyspark.sql import types as T

import laktory

schema = T.StructType(
    [
        T.StructField("x@x", T.IntegerType()),
        T.StructField("y", T.ArrayType(T.StringType())),
        T.StructField(
            "z",
            T.StructType(
                [
                    T.StructField("id", T.IntegerType()),
                    T.StructField("email", T.StringType()),
                ]
            ),
        ),
        T.StructField(
            "u",
            T.ArrayType(
                T.StructType(
                    [
                        T.StructField("a", T.IntegerType()),
                        T.StructField("b", T.IntegerType()),
                    ]
                )
            ),
        ),
        T.StructField(
            "u2",
            T.ArrayType(
                T.StructType(
                    [
                        T.StructField("a", T.IntegerType()),
                        T.StructField("b", T.IntegerType()),
                    ]
                )
            ),
        ),
    ]
)

data = [
    (
        1,
        ["a", "b"],
        {"id": 3, "email": "@gmail.com"},
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
    ),
    (
        2,
        ["b", "c"],
        {"id": 2, "email": "@gmail.com"},
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
    ),
    (
        3,
        ["c", "d"],
        {"id": 1, "email": "@gmail.com"},
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
    ),
]

spark = SparkSession.builder.appName("UnitTesting").getOrCreate()

df = spark.createDataFrame(data, schema=schema)


def test_df_schema_flat():
    df.show()
    df.printSchema()
    schema = df.schema_flat()
    assert schema == [
        "x@x",
        "y",
        "z",
        "z.id",
        "z.email",
        "u",
        "u[*].a",
        "u[*].b",
        "u2",
        "u2[*].a",
        "u2[*].b",
    ]


def test_df_has_column():
    assert df.has_column("x@x")
    assert df.has_column("`x@x`")
    assert df.has_column("u[0].a")


if __name__ == "__main__":
    test_df_schema_flat()
    test_df_has_column()



import pulumi
import yaml
import json
import os
import re
import inflect
from typing import Any
from typing import TypeVar
from typing import TextIO
from pydantic import BaseModel as _BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import model_serializer
from laktory._settings import settings
from laktory._parsers import _snake_to_camel

Model = TypeVar("Model", bound="BaseModel")


def is_pattern(s):
    return r"\$\{" in s


class BaseModel(_BaseModel):
    """
    Parent class for all Laktory models offering generic functions and
    properties. This `BaseModel` class is derived from `pydantic.BaseModel`.

    Attributes
    ----------
    variables:
        Variable values to be resolved when using `inject_vars` method.
    """

    model_config = ConfigDict(extra="forbid")
    variables: dict[str, Any] = Field(default={}, exclude=True)

    @model_serializer(mode="wrap")
    def camel_serializer(self, handler) -> dict[str, Any]:
        dump = handler(self)
        if dump is None:
            return dump

        if settings.camel_serialization:
            keys = list(dump.keys())
            for k in keys:
                k_camel = _snake_to_camel(k)
                if k_camel != k:
                    dump[_snake_to_camel(k)] = dump.pop(k)

        if settings.singular_serialization:
            engine = inflect.engine()
            fields = self.model_fields
            keys = list(dump.keys())
            for k in keys:
                if k in self.singularizations:
                    # Explicit singularization
                    k_singular = self.singularizations[k] or k
                else:
                    # Automatic singularization
                    k_singular = k
                    ann = str(fields[k].annotation)
                    if ann.startswith("list[laktory.models"):
                        k_singular = engine.singular_noun(k) or k

                if k_singular != k:
                    dump[k_singular] = dump.pop(k)

        return dump

    @classmethod
    def model_validate_yaml(cls, fp: TextIO) -> Model:
        """
        Load model from yaml file object. Other yaml files can be referenced
        using the ${include.other_yaml_filepath} syntax.

        Parameters
        ----------
        fp:
            file object structured as a yaml file

        Returns
        -------
        :
            Model instance
        """

        if hasattr(fp, "name"):
            dirpath = os.path.dirname(fp.name)
        else:
            dirpath = "./"

        def inject_includes(d):
            if isinstance(d, dict):
                for key, value in d.items():
                    d[key] = inject_includes(value)
            elif isinstance(d, list):
                for i, item in enumerate(d):
                    d[i] = inject_includes(item)
            elif "${include." in str(d):
                path = d.replace("${include.", "")[:-1]
                if not os.path.isabs(path):
                    path = os.path.join(dirpath, path)
                with open(path, "r") as _fp:
                    d = yaml.safe_load(_fp)
                    d = inject_includes(d)
            return d

        data = inject_includes(yaml.safe_load(fp))

        return cls.model_validate(data)

    @classmethod
    def model_validate_json_file(cls, fp: TextIO) -> Model:
        """
        Load model from json file object

        Parameters
        ----------
        fp:
            file object structured as a json file

        Returns
        -------
        :
            Model instance
        """
        data = json.load(fp)
        return cls.model_validate(data)

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {}

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def inject_vars(self, d: dict) -> dict[str, Any]:
        """
        Inject variables values into a dictionary (generally model dump).

        There are 3 types of variables:

        - User defined variables expressed as `${vars.variable_name}` and
          defined in `self.variables` or in environment variables.
        - Pulumi resources expressed as `${resources.resource_name}`. These
          are available from `laktory.pulumi_resources` and are populated
          automatically by Laktory.
        - Pulumi resources output properties expressed as
         `${resources.resource_name.output}`. These are available from
         `laktory.pulumi_outputs` and are populated automatically by
          Laktory.

        Pulumi Outputs are also supported as variable values.

        Parameters
        ----------
        d:
            Model dump

        Returns
        -------
        :
            Dump in which variable expressions have been replaced with their
            values.
        """

        from laktory.models.resources.pulumiresource import pulumi_outputs
        from laktory.models.resources.pulumiresource import pulumi_resources

        # Build patterns
        _patterns = {}
        _vars = {}
        _pvars = {}

        # User-defined variables
        for k, v in self.variables.items():
            _k = k
            if not is_pattern(_k):
                _k = f"${{vars.{_k}}}"
            if isinstance(v, pulumi.Output):
                _vars[_k] = f"{{_pargs_{k}}}"
                _pvars[f"_pargs_{k}"] = v
            else:
                _vars[_k] = v

        # Environment variables
        for k, v in os.environ.items():
            _vars[f"${{vars.{k}}}"] = v

        # Pulumi resource outputs
        for k, v in pulumi_outputs.items():
            _vars[f"${{resources.{k}}}"] = v

        # Pulumi resources
        for k, v in pulumi_resources.items():
            _vars[f"${{resources.{k}}}"] = v

        # Create patterns
        keys = list(_vars.keys())
        for k in keys:
            v = _vars[k]
            if isinstance(v, str) and not is_pattern(k):
                pattern = re.escape(k)
                pattern = rf"{pattern}"
                _vars[pattern] = _vars.pop(k)

        def search_and_replace(d, pattern, repl):
            if isinstance(d, dict):
                for key, value in d.items():
                    # if isinstance(key, str) and re.findall(pattern, key, flags=re.IGNORECASE):
                    #     k2 = re.sub(pattern, repl, key, flags=re.IGNORECASE)
                    #     d[k2] = search_and_replace(value, pattern, repl)
                    #     if key != k2:
                    #         del d[key]
                    # else:
                    d[key] = search_and_replace(value, pattern, repl)
            elif isinstance(d, list):
                for i, item in enumerate(d):
                    d[i] = search_and_replace(item, pattern, repl)
            elif (
                d == pattern
            ):  # required where d is not a string (bool or resource object)
                d = repl
            elif isinstance(d, str) and re.findall(pattern, d, flags=re.IGNORECASE):
                d = re.sub(pattern, repl, d, flags=re.IGNORECASE)

            return d

        def apply_pulumi(d):
            if isinstance(d, dict):
                for key, value in d.items():
                    d[key] = apply_pulumi(value)
            elif isinstance(d, list):
                for i, item in enumerate(d):
                    d[i] = apply_pulumi(item)
            elif isinstance(d, str) and "_pargs_" in d:
                d = pulumi.Output.all(**_pvars).apply(
                    lambda args, _d=d: _d.format_map(args)
                )
            else:
                pass
            return d

        # Replace variable with their values (except for pulumi output)
        for pattern, repl in _vars.items():
            d = search_and_replace(d, pattern, repl)

        # Build pulumi output function where required
        d = apply_pulumi(d)

        return d



import pyspark.sql.functions as F

from pyspark.sql.column import Column
from typing import Any
from laktory.spark.functions._common import (
    COLUMN_OR_NAME,
    INT_OR_COLUMN,
    FLOAT_OR_COLUMN,
    STRING_OR_COLUMN,
    _col,
    _lit,
)

__all__ = [
    "compare",
]


# --------------------------------------------------------------------------- #
# compare                                                                     #
# --------------------------------------------------------------------------- #


def compare(
    x: COLUMN_OR_NAME,
    y: COLUMN_OR_NAME = 0,
    where: COLUMN_OR_NAME = None,
    operator: str = "==",
    default: COLUMN_OR_NAME = None,
) -> Column:
    """
    Compare a column `x` and a value or another column `y` using
    `operator`. Comparison can be limited to `where` and assigned
    `default` elsewhere.

    output = `x` `operator` `y`

    Parameters
    ---------
    x :
        Base column to compare
    y :
        Column to compare to
    where:
        Where to apply the comparison
    operator: str
        Operator for comparison
    default:
        Default value to be applied when `where` is `False`

    Returns
    -------
    :
        Comparison result

    Examples
    --------
    ```py
    import pyspark.sql.functions as F
    import laktory.spark.functions as LF

    df = spark.createDataFrame([[0.45], [0.55]], ["x"])
    df = df.withColumn(
        "y",
        LF.compare(
            "x",
            F.lit(0.5),
            operator=">",
        ),
    )
    print(df.show_string())
    '''
    +----+-----+
    |   x|    y|
    +----+-----+
    |0.45|false|
    |0.55| true|
    +----+-----+
    '''
    ```
    """

    x = _col(x)
    y = _col(y)

    if operator == "==":
        c = x == y
    elif operator == "!=":
        c = x != y
    elif operator == "<":
        c = x < y
    elif operator == "<=":
        c = x <= y
    elif operator == ">":
        c = x > y
    elif operator == ">=":
        c = x >= y
    else:
        raise ValueError(f"Operator '{operator}' is not supported.")

    if where is not None:
        where = _col(where)
        if default is None:
            default = F.lit(None)
        else:
            default = _col(default)

        c = F.when(where, c).otherwise(default)

    return c



import pytest
from pytest_examples import find_examples, CodeExample, EvalExample
from pyspark.sql import SparkSession

spark = SparkSession.builder.getOrCreate()


# --------------------------------------------------------------------------- #
# DLT                                                                         #
# --------------------------------------------------------------------------- #


@pytest.mark.parametrize("example", find_examples("./laktory/dlt"), ids=str)
def test_docstrings_dlt(example: CodeExample, eval_example: EvalExample):
    if eval_example.update_examples:
        eval_example.format(example)
        eval_example.run_print_update(
            example,
            module_globals={
                "spark": spark,
                "display": lambda x: x,
            },
        )
    else:
        eval_example.lint(example)
        eval_example.run_print_check(
            example,
            module_globals={
                "spark": spark,
                "display": lambda x: x,
            },
        )


# --------------------------------------------------------------------------- #
# Models                                                                      #
# --------------------------------------------------------------------------- #


@pytest.mark.parametrize("example", find_examples("./laktory/models"), ids=str)
def test_docstrings_models(example: CodeExample, eval_example: EvalExample):
    if eval_example.update_examples:
        eval_example.format(example)
        eval_example.run_print_update(
            example,
            module_globals={"spark": spark},
        )
    else:
        eval_example.lint(example)
        eval_example.run_print_check(example, module_globals={"spark": spark})


# --------------------------------------------------------------------------- #
# Spark DataFrame                                                             #
# --------------------------------------------------------------------------- #


@pytest.mark.parametrize("example", find_examples("./laktory/spark/dataframe"), ids=str)
def test_docstrings_spark_dataframe(example: CodeExample, eval_example: EvalExample):
    if eval_example.update_examples:
        eval_example.format(example)
        eval_example.run_print_update(example, module_globals={"spark": spark})
    else:
        eval_example.lint(example)
        eval_example.run_print_check(example, module_globals={"spark": spark})


# --------------------------------------------------------------------------- #
# Spark Functions                                                             #
# --------------------------------------------------------------------------- #


@pytest.mark.parametrize("example", find_examples("./laktory/spark/functions"), ids=str)
def test_docstrings_spark_functions(example: CodeExample, eval_example: EvalExample):
    if eval_example.update_examples:
        eval_example.format(example)
        eval_example.run_print_update(example, module_globals={"spark": spark})
    else:
        eval_example.lint(example)
        eval_example.run_print_check(example, module_globals={"spark": spark})


# --------------------------------------------------------------------------- #
# Markdowns                                                                   #
# --------------------------------------------------------------------------- #


@pytest.mark.parametrize("example", find_examples("./docs/"), ids=str)
def test_docstrings_markdowns(example: CodeExample, eval_example: EvalExample):
    """
    Examples in documentation are not all designed to be run in isolation. We only format them using black. To be
    investigated if there is a better approach.
    """

    if eval_example.update_examples:
        eval_example.format_black(example)
    else:
        eval_example.lint_black(example)



from pyspark.sql import types as T
from pyspark.sql import functions as F
from pyspark.sql import SparkSession
import pandas as pd
import datetime

from laktory._testing import EventsManager
from laktory._testing import df_meta
from laktory._testing import table_brz
from laktory._testing import table_slv
from laktory._testing import table_slv_star
from laktory._testing import table_gld
from laktory.models import Table
from laktory.models import TableJoin
from laktory.models import TableWindowFilter
from laktory.models import TableAggregation

manager = EventsManager()
manager.build_events()

# Spark
spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
spark.conf.set("spark.sql.session.timeZone", "UTC")


def test_table_join():
    join = TableJoin(
        left={
            "name": "slv_stock_prices",
            "filter": "created_at = '2023-11-01T00:00:00Z'",
        },
        other={
            "name": "slv_stock_metadata",
            "selects": {
                "symbol2": "symbol",
                "currency": "currency",
                "first_traded": "first_traded",
            },
        },
        on=["symbol"],
    )
    join.left._df = table_slv.to_df(spark)
    join.other._df = df_meta

    df = join.execute(spark)
    data = df.toPandas().to_dict(orient="records")
    print(data)
    assert df.select("symbol")  # test for duplicated columns
    assert data == [
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "AAPL",
            "open": 1,
            "close": 2,
            "currency": "USD",
            "first_traded": "1980-12-12T14:30:00.000Z",
        },
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "GOOGL",
            "open": 3,
            "close": 4,
            "currency": "USD",
            "first_traded": "2004-08-19T13:30:00.00Z",
        },
    ]


def test_table_join_expression():
    join = TableJoin(
        left={
            "name": "slv_stock_prices",
            "filter": "created_at = '2023-11-01T00:00:00Z'",
        },
        other={
            "name": "slv_stock_metadata",
            "selects": {
                "symbol2": "symbol",
                "currency": "currency",
                "first_traded": "first_traded",
            },
        },
        on_expression="left.symbol == other.symbol",
    )
    join.left._df = table_slv.to_df(spark)
    join.other._df = df_meta

    df = join.execute(spark)
    data = df.toPandas().to_dict(orient="records")
    print(data)
    assert df.select("symbol")  # test for duplicated columns
    assert data == [
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "AAPL",
            "open": 1,
            "close": 2,
            "currency": "USD",
            "first_traded": "1980-12-12T14:30:00.000Z",
        },
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "GOOGL",
            "open": 3,
            "close": 4,
            "currency": "USD",
            "first_traded": "2004-08-19T13:30:00.00Z",
        },
    ]


def test_table_join_outer():
    join = TableJoin(
        left={
            "name": "slv_stock_prices",
            "filter": "created_at = '2023-11-01T00:00:00Z'",
        },
        other={
            "name": "slv_stock_metadata",
            "selects": {
                "symbol2": "symbol",
                "currency": "currency",
                "first_traded": "first_traded",
            },
        },
        on=["symbol"],
        how="full_outer",
    )
    join.left._df = table_slv.to_df(spark)
    join.other._df = df_meta

    df = join.execute(spark)

    # Test join columns uniqueness
    _df = df.withColumn("symbol2", F.lit("a"))
    # _df = df.withColumn("symbol", F.lit("a"))
    _df = _df.select("symbol")

    # Test data
    data = df.toPandas().fillna(-1).to_dict(orient="records")
    print(data)
    assert data == [
        {
            "created_at": "2023-11-01T00:00:00Z",
            "open": 1.0,
            "close": 2.0,
            "currency": "USD",
            "first_traded": "1980-12-12T14:30:00.000Z",
            "symbol": "AAPL",
        },
        {
            "created_at": -1,
            "open": -1.0,
            "close": -1.0,
            "currency": "USD",
            "first_traded": "1997-05-15T13:30:00.000Z",
            "symbol": "AMZN",
        },
        {
            "created_at": "2023-11-01T00:00:00Z",
            "open": 3.0,
            "close": 4.0,
            "currency": "USD",
            "first_traded": "2004-08-19T13:30:00.00Z",
            "symbol": "GOOGL",
        },
    ]


def test_table_agg():
    agg = TableAggregation(
        groupby_columns=["symbol"],
        agg_expressions=[
            {"name": "min_open", "spark_func_name": "min", "spark_func_args": ["open"]},
            {
                "name": "max_open",
                "sql_expression": "max(open)",
            },
        ],
    )
    df = table_slv.to_df(spark)
    df1 = agg.execute(df=df)
    df1.show()
    data = df1.toPandas().to_dict(orient="records")
    print(data)
    assert data == [
        {"symbol": "AAPL", "min_open": 1, "max_open": 3},
        {"symbol": "GOOGL", "min_open": 3, "max_open": 5},
    ]


def test_table_agg_window():
    agg = TableAggregation(
        groupby_window={
            "time_column": "created_at",
            "window_duration": "1 day",
            # "start_time": "2000-01-01T00:00:00Z"
        },
        groupby_columns=["symbol"],
        agg_expressions=[
            {"name": "min_open", "spark_func_name": "min", "spark_func_args": ["open"]},
            {
                "name": "max_open",
                "sql_expression": "max(open)",
            },
        ],
    )
    df = table_slv.to_df(spark)
    assert df.count() == 4

    df1 = agg.execute(df=df)
    df1.select("window.start", "window.end").show()
    df1.printSchema()

    # Convert window to string for easier testing
    df1 = df1.withColumn("start", F.col("window.start").cast("string"))
    df1 = df1.withColumn("end", F.col("window.end").cast("string"))
    df1 = df1.drop("window")
    df1.printSchema()

    data = [row.asDict(recursive=True) for row in df1.collect()]
    print(data)
    assert data == [
        {
            "symbol": "AAPL",
            "min_open": 1,
            "max_open": 3,
            "start": "2023-11-01 00:00:00",
            "end": "2023-11-02 00:00:00",
        },
        {
            "symbol": "GOOGL",
            "min_open": 3,
            "max_open": 5,
            "start": "2023-11-01 00:00:00",
            "end": "2023-11-02 00:00:00",
        },
    ]


def test_table_window_filter():
    w = TableWindowFilter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=1,
    )
    df = table_slv.to_df(spark)
    df.show()
    df1 = w.execute(df=df)
    df1.show()
    data = df1.toPandas().to_dict(orient="records")
    print(data)
    assert data == [
        {
            "created_at": "2023-11-01T01:00:00Z",
            "symbol": "AAPL",
            "open": 3,
            "close": 4,
            "_row_index": 1,
        },
        {
            "created_at": "2023-11-01T01:00:00Z",
            "symbol": "GOOGL",
            "open": 5,
            "close": 6,
            "_row_index": 1,
        },
    ]


def test_bronze():
    df0 = manager.to_spark_df()
    df1 = table_brz.builder.process(df0)
    assert "_bronze_at" in df1.columns


def test_silver():
    df0 = manager.to_spark_df()
    df0.show()
    df1 = table_brz.builder.process(df0)
    df1.show()
    df2 = table_slv.builder.process(df1)
    df2.show()
    assert df2.schema == T.StructType(
        [
            T.StructField("created_at", T.TimestampType(), True),
            T.StructField("symbol", T.StringType(), True),
            T.StructField("open", T.DoubleType(), True),
            T.StructField("close", T.DoubleType(), True),
            T.StructField("_bronze_at", T.TimestampType(), False),
            T.StructField("_silver_at", T.TimestampType(), False),
        ]
    )
    s = df2.toPandas().iloc[0]
    print(s)
    assert s["created_at"] == pd.Timestamp("2023-09-01 00:00:00")
    assert s["symbol"] == "AAPL"
    assert s["open"] == 189.49000549316406
    assert s["close"] == 189.49000549316406


def test_silver_star():
    df = table_slv_star.builder.read_source(spark)
    df = table_slv_star.builder.process(df, spark=spark)
    df.printSchema()

    assert "_silver_star_at" in df.columns
    data = df.toPandas().drop("_silver_star_at", axis=1).to_dict("records")
    print(data)
    assert data == [
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "AAPL",
            "open": 1,
            "close": 2,
            "symbol3": "AAPL",
            "currency": "USD",
            "last_traded": "1980-12-12T14:30:00.000Z",
            "name": "Apple",
        },
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "GOOGL",
            "open": 3,
            "close": 4,
            "symbol3": "GOOGL",
            "currency": "USD",
            "last_traded": "2004-08-19T13:30:00.00Z",
            "name": "Google",
        },
    ]


def test_gold():
    assert table_gld.builder.template == "GOLD1"
    df0 = manager.to_spark_df()
    df0.show()
    df1 = table_brz.builder.process(df0)
    df1.show()
    df2 = table_slv.builder.process(df1)
    df2.show()
    df3 = table_gld.builder.process(df2)
    df3.show()
    df3.printSchema()
    assert df3.schema == T.StructType(
        [
            T.StructField("min_open", T.DoubleType(), True),
            T.StructField("max_open", T.DoubleType(), True),
            T.StructField("min_close", T.DoubleType(), True),
            T.StructField("_gold_at", T.TimestampType(), False),
            T.StructField("name", T.StringType(), True),
            T.StructField("symbol", T.StringType(), True),
            T.StructField("name2", T.StringType(), True),
        ]
    )
    s = df3.toPandas().iloc[0]
    print(s)
    assert s["name2"] == "Apple"


def test_drop_duplicates():
    # Build DataFrame with duplicates
    df0 = manager.to_spark_df()
    df0.show()
    df1 = table_brz.builder.process(df0)
    df1.show()
    df1 = df1.union(df1)

    # Remove duplicates using all columns
    slv_tmp = table_slv.copy()
    slv_tmp.builder.drop_duplicates = True
    df2 = slv_tmp.builder.process(df1)
    df2.show()

    # Remove duplicates using only symbol and _bronze_at
    slv_tmp.builder.drop_duplicates = ["symbol", "_bronze_at"]
    df3 = slv_tmp.builder.process(df1)

    assert df1.count() == 160
    assert df2.count() == 80
    assert df3.count() == 4


def test_cdc():
    table = Table(
        name="brz_users_type1",
        builder={
            "table_source": {
                "name": "brz_users_cdc",
                "cdc": {
                    "primary_keys": ["userId"],
                    "sequence_by": "sequenceNum",
                    "apply_as_deletes": "operation = 'DELETE'",
                    "scd_type": 1,
                    "except_columns": ["operation", "sequenceNum"],
                },
            },
        },
    )

    assert table.builder.apply_changes_kwargs == {
        "apply_as_deletes": "operation = 'DELETE'",
        "apply_as_truncates": None,
        "column_list": [],
        "except_column_list": ["operation", "sequenceNum"],
        "ignore_null_updates": None,
        "keys": ["userId"],
        "sequence_by": "sequenceNum",
        "source": "brz_users_cdc",
        "stored_as_scd_type": 1,
        "target": "brz_users_type1",
        "track_history_column_list": None,
        "track_history_except_column_list": None,
    }
    assert table.builder.is_from_cdc

    # TODO: Run test with demo data
    # from pyspark.sql import SparkSession

    # spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
    #
    # df_cdc = spark.createDataFrame(pd.DataFrame({
    #     "userId": [124, 123, 125, 126, 123, 125, 125, 123],
    #     "name": ["Raul", "Isabel", "Mercedes", "Lily", None, "Mercedes", "Mercedes", "Isabel"],
    #     "city": ["Oaxaca", "Monterrey", "Tijuana", "Cancun", None, "Guadalajara", "Mexicali", "Chihuahua"],
    #     "operation": ["INSERT", "INSERT", "INSERT", "INSERT", "DELETE", "UPDATE", "UPDATE", "UPDATE"],
    #     "sequenceNum": [1, 1, 2, 2, 6, 6, 5, 5],
    # }))
    # df_cdc.show()


if __name__ == "__main__":
    test_table_join()
    test_table_join_expression()
    test_table_join_outer()
    test_table_agg()
    test_table_agg_window()
    test_table_window_filter()
    test_bronze()
    test_silver()
    test_silver_star()
    test_gold()
    test_cdc()
    test_drop_duplicates()



from typing import Literal
from typing import Union
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.accesscontrol import AccessControl
from laktory.models.databricks.permissions import Permissions


class WarehouseCustomTag(BaseModel):
    """
    Warehouse Custom Tag specifications

    Attributes
    ----------
    key:
        Tag key
    value:
        Tag value
    """

    key: str
    value: str


class WarehouseTags(BaseModel):
    """
    Warehouse Tags specifications

    Attributes
    ----------
    custom_tags:
        Tags specifications
    """

    custom_tags: list[WarehouseCustomTag] = []


class Warehouse(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Warehouse

    Attributes
    ----------
    access_controls:
        Warehouse access controls
    auto_stop_mins:
        Time in minutes until an idle SQL warehouse terminates all clusters and stops.
    channel_name:
        Channel specifications
    cluster_size:
        The size of the clusters allocated to the endpoint
    enable_photon:
        If `True`, photon is enabled
    enable_serverless_compute:
        If `True`, warehouse is serverless.
        See details [here](https://www.pulumi.com/registry/packages/databricks/api-docs/sqlendpoint/)
    instance_profile_arn:
        TODO
    jdbc_url:
        JDBC connection string.
    max_num_clusters:
        Maximum number of clusters available when a SQL warehouse is running.
    min_num_clusters:
        Minimum number of clusters available when a SQL warehouse is running.
    name:
        Warehouse name
    num_clusters:
        Fixed number of clusters when autoscaling is not enabled.
    spot_instance_policy:
        The spot policy to use for allocating instances to clusters.
    tags:
        Databricks tags all endpoint resources with these tags.
    warehouse_type:
        SQL warehouse type.

    Examples
    --------
    ```py
    from laktory import models

    warehouse = models.Warehouse(
        name="default",
        cluster_size="2X-Small",
        auto_stop_mins=30,
        channel_name="CHANNEL_NAME_PREVIEW",
        enable_photon=True,
        enable_serverless_compute=True,
        access_controls=[{"group_name": "account users", "permission_level": "CAN_USE"}],
    )
    ```
    """

    cluster_size: Literal[
        "2X-Small",
        "X-Small",
        "Small",
        "Medium",
        "Large",
        "X-Large",
        "2X-Large",
        "3X-Large",
        "4X-Large",
    ]
    access_controls: list[AccessControl] = []
    auto_stop_mins: int = None
    channel_name: Union[
        Literal["CHANNEL_NAME_CURRENT", "CHANNEL_NAME_PREVIEW"], str
    ] = None
    # data_source_id
    enable_photon: bool = None
    enable_serverless_compute: bool = None
    instance_profile_arn: str = None
    jdbc_url: str = None
    max_num_clusters: int = None
    min_num_clusters: int = None
    name: str = None
    num_clusters: int = None
    # odbc_params
    spot_instance_policy: Union[
        Literal["COST_OPTIMIZED", "RELIABILITY_OPTIMIZED"], str
    ] = None
    # state
    tags: WarehouseTags = None
    warehouse_type: Union[Literal["CLASSIC", "PRO"], str] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - warehouse permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    sql_endpoint_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SqlEndpoint"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.SqlEndpoint

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        if settings.camel_serialization:
            d["channel"] = {"name": d.pop("channelName")}
        else:
            d["channel"] = {"name": d.pop("channel_name")}
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_sql_endpoint"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        d["channel"] = {"name": d.pop("channel_name")}
        return d



import re
from typing import Any
from pydantic import AliasChoices
from pydantic import Field
from pydantic import BaseModel as _BaseModel
from laktory.models.basemodel import BaseModel


class ResourceOptions(BaseModel):
    """
    Resource options for deployment.

    Attributes
    ----------
    depends_on:
        Explicit list of resources dependencies.
        Supported by both pulumi and terraform.
    provider:
        Explicit declaration of resources provider.
        Supported by both pulumi and terraform.
    aliases:
        Specify aliases for this resource, so that renaming or refactoring
        doesn’t replace it.
        Pulumi only.
    delete_before_replace:
        Override the default create-before-delete behavior when replacing a
        resource.
        Pulumi only.
    ignore_changes:
        Declare that changes to certain properties should be ignored during a
        diff.
        Pulumi only.
    import_:
        Bring an existing cloud resource into Pulumi.
        Pulumi only.
    parent:
        Establish a parent/child relationship between resources.
        Pulumi only.
    replace_on_changes:
        Declare that changes to certain properties should be treated as forcing
        a replacement.
        Pulumi only.
    """

    # pulumi + terraform
    depends_on: list[str] = []
    provider: str = None

    # pulumi only
    aliases: list[str] = None
    delete_before_replace: bool = True
    ignore_changes: list[str] = None
    import_: str = None
    parent: str = None
    replace_on_changes: list[str] = None


class BaseResource(_BaseModel):
    """
    Parent class for all Laktory models deployable as one or multiple cloud
    core resources. This `BaseResource` class is derived from
    `pydantic.BaseModel`.

    Attributes
    ----------
    resource_name:
        Name of the resource in the context of infrastructure as code. If None,
        `default_resource_name` will be used instead.
    options:
        Resources options specifications
    """

    resource_name_: str = Field(
        None,
        validation_alias=AliasChoices("resource_name_", "resource_name"),
        exclude=True,
    )
    options: ResourceOptions = Field(ResourceOptions(), exclude=True)
    _core_resources: list[Any] = None

    @property
    def resource_name(self) -> str:
        pattern = re.compile(r"^[a-zA-Z][a-zA-Z0-9-_]*$")

        name = self.default_resource_name
        if self.resource_name_:
            name = self.resource_name_

        if not pattern.match(name):
            raise ValueError(
                f"Resource name `{name}` is invalid. A name must start with a letter or underscore and may contain only letters, digits, underscores, and dashes."
            )

        return name

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        """
        Resource type id used to build default resource name. Equivalent to
        class name converted to kebab case. Ex.: SecretScope -> secret-scope
        """
        _id = type(self).__name__
        _id = re.sub(
            r"(?<!^)(?=[A-Z])", "-", _id
        ).lower()  # Convert CamelCase to kebab-case
        return _id

    @property
    def resource_key(self) -> str:
        """
        Resource key used to build default resource name. Equivalent to
        name properties if available. Otherwise, empty string.
        """
        return getattr(self, "name", "")

    @property
    def default_resource_name(self) -> str:
        """
        Resource default name constructed as
        - `{self.resource_type_id}-{self.resource_key}`
        - removing ${resources....} tags
        - removing ${vars....} tags
        - Replacing special characters with - to avoid conflicts with resource properties
        """

        if self.resource_type_id not in self.resource_key:
            name = f"{self.resource_type_id}-{self.resource_key}"
        else:
            name = f"{self.resource_key}"

        if name.endswith("-"):
            name = name[:-1]

        # ${resources.x.property} -> x
        pattern = r"\$\{resources\.(.*?)\.(.*?)\}"
        name = re.sub(pattern, r"\1", name)

        # ${vars.x} -> x
        pattern = r"\$\{vars\.(.*?)\}"
        name = re.sub(pattern, r"\1", name)

        # Replace special characters
        chars = [".", "@"]
        for c in chars:
            name = name.replace(c, "-")

        return name

    @property
    def self_as_core_resources(self):
        """Flag set to `True` if self must be included in core resources"""
        return True

    @property
    def additional_core_resources(self):
        return []

    @property
    def core_resources(self):
        """
        List of core resources to be deployed with this laktory model:
        - class instance (self)
        """
        if self._core_resources is None:
            # Get all resources
            self._core_resources = []
            if self.self_as_core_resources:
                self._core_resources += [self]
            self._core_resources += self.additional_core_resources

            # Propagate options
            r0 = self._core_resources[0]
            provider = r0.options.provider
            k0 = f"${{resources.{r0.resource_name}}}"
            for r in self._core_resources[1:]:
                if provider:
                    if r.options.provider is None:
                        r.options.provider = provider

                do = r.options.depends_on
                if k0 not in do:
                    do += [k0]
                r.options.depends_on = do

        return self._core_resources



from laktory.models import WorkspaceFile

workspace_file = WorkspaceFile(
    source="./test_workspacefile.py",
    dirpath="/init_scripts/",
    access_controls=[{"permission_level": "CAN_READ", "group_name": "account users"}],
)


def test_workspace_file():
    print(workspace_file)
    assert workspace_file.filename == "test_workspacefile.py"
    assert workspace_file.path == "/init_scripts/test_workspacefile.py"
    assert workspace_file.resource_key == "init_scripts-test_workspacefile-py"
    assert (
        workspace_file.resource_name
        == "workspace-file-init_scripts-test_workspacefile-py"
    )

    assert workspace_file.access_controls[0].permission_level == "CAN_READ"
    assert workspace_file.access_controls[0].group_name == "account users"


if __name__ == "__main__":
    test_workspace_file()



from datetime import datetime
from zoneinfo import ZoneInfo
import pytest

from laktory._testing import StockPriceDataEventHeader
from laktory._testing import EventsManager
from laktory import models
from laktory import settings

header = StockPriceDataEventHeader()
events = EventsManager().build_events()
event = events[0]


def test_dataeventheader():
    print(header.model_dump())
    assert header.model_dump() == {
        "name": "stock_price",
        "description": None,
        "producer": {"name": "yahoo-finance", "description": None, "party": 1},
        "events_root_": None,
        "event_root_": None,
    }
    assert header.events_root == "/Volumes/dev/sources/landing/events/"
    assert (
        header.event_root
        == "/Volumes/dev/sources/landing/events/yahoo-finance/stock_price/"
    )


def test_dataeventheader_root():
    # Custom value for events root
    header = models.DataEventHeader(name="my-event", events_root="test")
    header2 = models.DataEventHeader(
        **header.model_dump(by_alias=True, exclude_none=True)
    )
    assert header.events_root == "test"
    assert header2.events_root == "test"

    # Default value
    header = models.DataEventHeader(name="my-event")
    assert header.events_root == "/Volumes/dev/sources/landing/events/"
    root0 = settings.workspace_landing_root
    header2 = models.DataEventHeader(
        **header.model_dump(by_alias=True, exclude_none=True)
    )
    settings.workspace_landing_root = "custom_root/"
    assert header.events_root == "custom_root/events/"
    assert header2.events_root == "custom_root/events/"
    settings.workspace_landing_root = root0

    # Custom value for event root
    header = models.DataEventHeader(name="my-event", event_root="/Landing/my_event")
    print(header)
    assert header.event_root == "/Landing/my_event"
    header = models.DataEventHeader(**header.model_dump(exclude_none=True))


def test_dataevent():
    assert event.producer.name == "yahoo-finance"
    assert event.data["symbol"] == "AAPL"
    assert event.data["open"] == pytest.approx(189.49, abs=0.01)
    assert event.data["_producer_name"] == "yahoo-finance"
    assert event.data["_created_at"] == datetime(2023, 9, 1, tzinfo=ZoneInfo("UTC"))
    assert event.created_at == datetime(2023, 9, 1, 0, 0, 0, tzinfo=ZoneInfo("UTC"))
    assert (
        event.get_landing_filepath()
        == "/Volumes/dev/sources/landing/events/yahoo-finance/stock_price/2023/09/01/stock_price_20230901T000000000Z.json"
    )


def test_model_dump():
    # Without exclusions
    d = event.model_dump(exclude=[])
    assert d == {
        "name": "stock_price",
        "description": None,
        "producer": {"name": "yahoo-finance", "description": None, "party": 1},
        "events_root": None,
        "event_root": None,
        "data": {
            "created_at": "2023-09-01T00:00:00",
            "symbol": "AAPL",
            "open": 189.49000549316406,
            "close": 189.4600067138672,
            "high": 189.9199981689453,
            "low": 188.27999877929688,
            "@id": "_id",
            "_name": "stock_price",
            "_producer_name": "yahoo-finance",
            "_created_at": "2023-09-01T00:00:00Z",
        },
        "tstamp_col": "created_at",
        "tstamp_in_path": True,
    }

    assert event.events_root == "/Volumes/dev/sources/landing/events/"

    # With exclusions
    d = event.model_dump()
    print(d)
    assert d == {
        "name": "stock_price",
        "description": None,
        "producer": {"name": "yahoo-finance", "description": None, "party": 1},
        "event_root": None,
        "data": {
            "created_at": "2023-09-01T00:00:00",
            "symbol": "AAPL",
            "open": 189.49000549316406,
            "close": 189.4600067138672,
            "high": 189.9199981689453,
            "low": 188.27999877929688,
            "@id": "_id",
            "_name": "stock_price",
            "_producer_name": "yahoo-finance",
            "_created_at": "2023-09-01T00:00:00Z",
        },
    }


def test_event_without_tstamp():
    d = event.model_dump()
    d["tstamp_in_path"] = False
    e = models.DataEvent(**d)
    assert (
        e.get_landing_filepath()
        == "/Volumes/dev/sources/landing/events/yahoo-finance/stock_price/stock_price.json"
    )


def test_to_azure_storage_container():
    try:
        import azure.storage
    except ModuleNotFoundError:
        return

    event.to_azure_storage_container(container_name="unit-testing", overwrite=True)
    with pytest.raises(FileExistsError):
        event.to_azure_storage_container(container_name="unit-testing")
    event.to_azure_storage_container(container_name="unit-testing", skip_if_exists=True)


def test_to_aws_s3_bucket():
    try:
        import boto3
    except ModuleNotFoundError:
        return

    event.to_aws_s3_bucket(bucket_name="okube-unit-testing", overwrite=True)
    with pytest.raises(FileExistsError):
        event.to_aws_s3_bucket(bucket_name="okube-unit-testing")
    event.to_aws_s3_bucket(bucket_name="okube-unit-testing", skip_if_exists=True)


def test_to_databricks_mount():
    # TODO: add tests
    pass
    # event.to_databricks_mount()


if __name__ == "__main__":
    test_dataeventheader()
    test_dataeventheader_root()
    test_dataevent()
    test_model_dump()
    test_event_without_tstamp()
    test_to_azure_storage_container()
    test_to_aws_s3_bucket()
    test_to_databricks_mount()



import copy
import os

from laktory._testing.stackvalidator import StackValidator
from laktory import models

dirpath = os.path.dirname(__file__)

with open(os.path.join(dirpath, "stack.yaml"), "r") as fp:
    stack = models.Stack.model_validate_yaml(fp)

stack.terraform.backend = {
    "azurerm": {
        "resource_group_name": "o3-rg-laktory-dev",
        "storage_account_name": "o3stglaktorydev",
        "container_name": "unit-testing",
        "key": "terraform/dev.terraform.tfstate",
    }
}


def test_stack_model():
    data = stack.model_dump()
    print(data)
    assert data == {
        "variables": {},
        "backend": "pulumi",
        "description": None,
        "environments": {
            "dev": {
                "variables": {
                    "env": "dev",
                    "is_dev": True,
                    "node_type_id": "Standard_DS3_v2",
                },
                "resources": None,
            },
            "prod": {
                "variables": {
                    "env": "prod",
                    "is_dev": False,
                    "node_type_id": "Standard_DS4_v2",
                },
                "resources": {"pipelines": {"pl-custom-name": {"development": False}}},
            },
        },
        "name": "unit-testing",
        "organization": "okube",
        "pulumi": {
            "config": {
                "databricks:host": "${vars.DATABRICKS_HOST}",
                "databricks:token": "${vars.DATABRICKS_TOKEN}",
            },
            "outputs": {},
        },
        "resources": {
            "catalogs": {},
            "clusters": {},
            "directories": {},
            "externallocations": {},
            "groups": {},
            "jobs": {
                "job-stock-prices-ut-stack": {
                    "access_controls": [],
                    "clusters": [
                        {
                            "apply_policy_default_values": None,
                            "autoscale": None,
                            "autotermination_minutes": None,
                            "cluster_id": None,
                            "custom_tags": None,
                            "data_security_mode": "USER_ISOLATION",
                            "driver_instance_pool_id": None,
                            "driver_node_type_id": None,
                            "enable_elastic_disk": None,
                            "enable_local_disk_encryption": None,
                            "idempotency_token": None,
                            "init_scripts": [],
                            "instance_pool_id": None,
                            "name": "main",
                            "node_type_id": "${vars.node_type_id}",
                            "num_workers": None,
                            "policy_id": None,
                            "runtime_engine": None,
                            "single_user_name": None,
                            "spark_conf": {},
                            "spark_env_vars": {
                                "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                            },
                            "spark_version": "14.0.x-scala2.12",
                            "ssh_public_keys": [],
                        }
                    ],
                    "continuous": None,
                    "control_run_state": None,
                    "email_notifications": None,
                    "format": None,
                    "health": None,
                    "max_concurrent_runs": None,
                    "max_retries": None,
                    "min_retry_interval_millis": None,
                    "name": "job-stock-prices-ut-stack",
                    "notification_settings": None,
                    "parameters": [],
                    "retry_on_timeout": None,
                    "run_as": None,
                    "schedule": None,
                    "tags": {},
                    "tasks": [
                        {
                            "condition_task": None,
                            "depends_ons": None,
                            "description": None,
                            "email_notifications": None,
                            "existing_cluster_id": None,
                            "health": None,
                            "job_cluster_key": "main",
                            "libraries": [
                                {
                                    "cran": None,
                                    "egg": None,
                                    "jar": None,
                                    "maven": None,
                                    "pypi": {
                                        "package": "laktory==0.0.27",
                                        "repo": None,
                                    },
                                    "whl": None,
                                },
                                {
                                    "cran": None,
                                    "egg": None,
                                    "jar": None,
                                    "maven": None,
                                    "pypi": {"package": "yfinance", "repo": None},
                                    "whl": None,
                                },
                            ],
                            "max_retries": None,
                            "min_retry_interval_millis": None,
                            "notebook_task": {
                                "notebook_path": "/jobs/ingest_stock_metadata.py",
                                "base_parameters": None,
                                "source": None,
                            },
                            "notification_settings": None,
                            "pipeline_task": None,
                            "retry_on_timeout": None,
                            "run_if": None,
                            "run_job_task": None,
                            "sql_task": None,
                            "task_key": "ingest-metadata",
                            "timeout_seconds": None,
                        },
                        {
                            "condition_task": None,
                            "depends_ons": None,
                            "description": None,
                            "email_notifications": None,
                            "existing_cluster_id": None,
                            "health": None,
                            "job_cluster_key": None,
                            "libraries": None,
                            "max_retries": None,
                            "min_retry_interval_millis": None,
                            "notebook_task": None,
                            "notification_settings": None,
                            "pipeline_task": {
                                "pipeline_id": "${resources.pl-custom-name.id}",
                                "full_refresh": None,
                            },
                            "retry_on_timeout": None,
                            "run_if": None,
                            "run_job_task": None,
                            "sql_task": None,
                            "task_key": "run-pipeline",
                            "timeout_seconds": None,
                        },
                    ],
                    "timeout_seconds": None,
                    "trigger": None,
                    "webhook_notifications": None,
                }
            },
            "metastoredataaccesses": {},
            "metastores": {},
            "notebooks": {},
            "pipelines": {
                "pl-custom-name": {
                    "access_controls": [
                        {
                            "group_name": "account users",
                            "permission_level": "CAN_VIEW",
                            "service_principal_name": None,
                            "user_name": None,
                        },
                        {
                            "group_name": "role-engineers",
                            "permission_level": "CAN_RUN",
                            "service_principal_name": None,
                            "user_name": None,
                        },
                    ],
                    "allow_duplicate_names": None,
                    "catalog": None,
                    "channel": "PREVIEW",
                    "clusters": [],
                    "configuration": {},
                    "continuous": None,
                    "development": None,
                    "edition": None,
                    "libraries": [
                        {
                            "file": None,
                            "notebook": {"path": "/pipelines/dlt_brz_template.py"},
                        }
                    ],
                    "name": "pl-stock-prices-ut-stack",
                    "notifications": [],
                    "photon": None,
                    "serverless": None,
                    "storage": None,
                    "tables": [],
                    "target": None,
                    "udfs": [],
                }
            },
            "schemas": {},
            "secrets": {},
            "secretscopes": {},
            "serviceprincipals": {},
            "sqlqueries": {},
            "tables": {},
            "providers": {
                "databricks": {
                    "account_id": None,
                    "auth_type": None,
                    "azure_client_id": None,
                    "azure_client_secret": None,
                    "azure_environment": None,
                    "azure_login_app_id": None,
                    "azure_tenant_id": None,
                    "azure_use_msi": None,
                    "azure_workspace_resource_id": None,
                    "client_id": None,
                    "client_secret": None,
                    "cluster_id": None,
                    "config_file": None,
                    "databricks_cli_path": None,
                    "debug_headers": None,
                    "debug_truncate_bytes": None,
                    "google_credentials": None,
                    "google_service_account": None,
                    "host": "${vars.DATABRICKS_HOST}",
                    "http_timeout_seconds": None,
                    "metadata_service_url": None,
                    "password": None,
                    "profile": None,
                    "rate_limit": None,
                    "retry_timeout_seconds": None,
                    "skip_verify": None,
                    "token": "${vars.DATABRICKS_TOKEN}",
                    "username": None,
                    "warehouse_id": None,
                }
            },
            "users": {},
            "volumes": {},
            "warehouses": {},
            "workspacefiles": {},
        },
        "terraform": {
            "backend": {
                "azurerm": {
                    "resource_group_name": "o3-rg-laktory-dev",
                    "storage_account_name": "o3stglaktorydev",
                    "container_name": "unit-testing",
                    "key": "terraform/dev.terraform.tfstate",
                }
            }
        },
    }

    return stack


def test_pulumi_stack():
    pstack = stack.to_pulumi(env=None)
    assert pstack.organization == "okube"
    data_default = pstack.model_dump()
    data_default["config"]["databricks:token"] = "***"
    data_default["resources"]["databricks"]["properties"]["token"] = "***"
    print(data_default)
    assert data_default == {
        "variables": {},
        "name": "unit-testing",
        "runtime": "yaml",
        "config": {
            "databricks:host": "https://adb-2211091707396001.1.azuredatabricks.net/",
            "databricks:token": "***",
        },
        "resources": {
            "job-stock-prices-ut-stack": {
                "type": "databricks:Job",
                "properties": {
                    "name": "job-stock-prices-ut-stack",
                    "parameters": [],
                    "tags": {},
                    "tasks": [
                        {
                            "libraries": [
                                {"pypi": {"package": "laktory==0.0.27"}},
                                {"pypi": {"package": "yfinance"}},
                            ],
                            "jobClusterKey": "main",
                            "notebookTask": {
                                "notebookPath": "/jobs/ingest_stock_metadata.py"
                            },
                            "taskKey": "ingest-metadata",
                        },
                        {
                            "pipelineTask": {"pipelineId": "${pl-custom-name.id}"},
                            "taskKey": "run-pipeline",
                        },
                    ],
                    "jobClusters": [
                        {
                            "jobClusterKey": "main",
                            "newCluster": {
                                "dataSecurityMode": "USER_ISOLATION",
                                "initScripts": [],
                                "nodeTypeId": "${vars.node_type_id}",
                                "sparkConf": {},
                                "sparkEnvVars": {
                                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                    "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                                },
                                "sparkVersion": "14.0.x-scala2.12",
                                "sshPublicKeys": [],
                            },
                        }
                    ],
                },
                "options": {"dependsOn": [], "deleteBeforeReplace": True},
            },
            "pl-custom-name": {
                "type": "databricks:Pipeline",
                "properties": {
                    "channel": "PREVIEW",
                    "clusters": [],
                    "configuration": {},
                    "libraries": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}}
                    ],
                    "name": "pl-stock-prices-ut-stack",
                    "notifications": [],
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": [],
                    "deleteBeforeReplace": True,
                },
            },
            "permissions-pl-custom-name": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "account users", "permissionLevel": "CAN_VIEW"},
                        {"groupName": "role-engineers", "permissionLevel": "CAN_RUN"},
                    ],
                    "pipelineId": "${pl-custom-name.id}",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": ["${pl-custom-name}"],
                    "deleteBeforeReplace": True,
                },
            },
            "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-json": {
                "type": "databricks:WorkspaceFile",
                "properties": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack.json",
                    "source": "./tmp-pl-stock-prices-ut-stack.json",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": ["${pl-custom-name}"],
                    "deleteBeforeReplace": True,
                },
            },
            "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-json": {
                "type": "databricks:Permissions",
                "properties": {
                    "accessControls": [
                        {"groupName": "account users", "permissionLevel": "CAN_READ"}
                    ],
                    "workspaceFilePath": "/.laktory/pipelines/pl-stock-prices-ut-stack.json",
                },
                "options": {
                    "provider": "${databricks}",
                    "dependsOn": [
                        "${workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-json}",
                        "${pl-custom-name}",
                    ],
                    "deleteBeforeReplace": True,
                },
            },
            "databricks": {
                "type": "pulumi:providers:databricks",
                "properties": {
                    "host": "https://adb-2211091707396001.1.azuredatabricks.net/",
                    "token": "***",
                },
                "options": {"dependsOn": [], "deleteBeforeReplace": True},
            },
        },
        "outputs": {},
    }

    # Dev
    data = stack.to_pulumi(env="dev").model_dump()
    data["config"]["databricks:token"] = "***"
    data["resources"]["databricks"]["properties"]["token"] = "***"
    data0 = copy.deepcopy(data_default)
    data0["variables"]["env"] = "dev"
    data0["variables"]["is_dev"] = True
    data0["variables"]["node_type_id"] = "Standard_DS3_v2"
    cluster = data0["resources"]["job-stock-prices-ut-stack"]["properties"][
        "jobClusters"
    ][0]["newCluster"]
    cluster["nodeTypeId"] = "Standard_DS3_v2"
    cluster["sparkEnvVars"]["LAKTORY_WORKSPACE_ENV"] = "dev"
    assert data == data0

    # Prod
    data = stack.to_pulumi(env="prod").model_dump()
    data["config"]["databricks:token"] = "***"
    data["resources"]["databricks"]["properties"]["token"] = "***"
    data0 = copy.deepcopy(data_default)
    data0["variables"]["env"] = "prod"
    data0["variables"]["is_dev"] = False
    data0["variables"]["node_type_id"] = "Standard_DS4_v2"
    cluster = data0["resources"]["job-stock-prices-ut-stack"]["properties"][
        "jobClusters"
    ][0]["newCluster"]
    cluster["nodeTypeId"] = "Standard_DS4_v2"
    cluster["sparkEnvVars"]["LAKTORY_WORKSPACE_ENV"] = "prod"
    data0["resources"]["pl-custom-name"]["properties"]["development"] = False
    assert data == data0


def test_pulumi_preview():
    pstack = stack.to_pulumi(env="dev")
    pstack.preview(stack="okube/dev")


def test_terraform_stack():
    data_default = stack.to_terraform().model_dump()
    data_default["provider"]["databricks"]["token"] = "***"
    print(data_default)
    assert data_default == {
        "terraform": {
            "required_providers": {"databricks": {"source": "databricks/databricks"}},
            "backend": {
                "azurerm": {
                    "resource_group_name": "o3-rg-laktory-dev",
                    "storage_account_name": "o3stglaktorydev",
                    "container_name": "unit-testing",
                    "key": "terraform/dev.terraform.tfstate",
                }
            },
        },
        "provider": {
            "databricks": {
                "host": "https://adb-2211091707396001.1.azuredatabricks.net/",
                "token": "***",
            }
        },
        "resource": {
            "databricks_job": {
                "job-stock-prices-ut-stack": {
                    "name": "job-stock-prices-ut-stack",
                    "tags": {},
                    "parameter": [],
                    "task": [
                        {
                            "job_cluster_key": "main",
                            "notebook_task": {
                                "notebook_path": "/jobs/ingest_stock_metadata.py"
                            },
                            "task_key": "ingest-metadata",
                            "library": [
                                {"pypi": {"package": "laktory==0.0.27"}},
                                {"pypi": {"package": "yfinance"}},
                            ],
                        },
                        {
                            "pipeline_task": {
                                "pipeline_id": "${databricks_pipeline.pl-custom-name.id}"
                            },
                            "task_key": "run-pipeline",
                        },
                    ],
                    "job_cluster": [
                        {
                            "job_cluster_key": "main",
                            "new_cluster": {
                                "data_security_mode": "USER_ISOLATION",
                                "init_scripts": [],
                                "node_type_id": "${vars.node_type_id}",
                                "spark_conf": {},
                                "spark_env_vars": {
                                    "AZURE_TENANT_ID": "{{secrets/azure/tenant-id}}",
                                    "LAKTORY_WORKSPACE_ENV": "${vars.env}",
                                },
                                "spark_version": "14.0.x-scala2.12",
                                "ssh_public_keys": [],
                            },
                        }
                    ],
                }
            },
            "databricks_pipeline": {
                "pl-custom-name": {
                    "channel": "PREVIEW",
                    "configuration": {},
                    "name": "pl-stock-prices-ut-stack",
                    "cluster": [],
                    "library": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}}
                    ],
                    "notification": [],
                    "provider": "databricks",
                }
            },
            "databricks_permissions": {
                "permissions-pl-custom-name": {
                    "pipeline_id": "${databricks_pipeline.pl-custom-name.id}",
                    "access_control": [
                        {"group_name": "account users", "permission_level": "CAN_VIEW"},
                        {"group_name": "role-engineers", "permission_level": "CAN_RUN"},
                    ],
                    "depends_on": ["databricks_pipeline.pl-custom-name"],
                    "provider": "databricks",
                },
                "permissions-workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-json": {
                    "workspace_file_path": "/.laktory/pipelines/pl-stock-prices-ut-stack.json",
                    "access_control": [
                        {"group_name": "account users", "permission_level": "CAN_READ"}
                    ],
                    "depends_on": [
                        "databricks_workspace_file.workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-json",
                        "databricks_pipeline.pl-custom-name",
                    ],
                    "provider": "databricks",
                },
            },
            "databricks_workspace_file": {
                "workspace-file-laktory-pipelines-pl-stock-prices-ut-stack-json": {
                    "path": "/.laktory/pipelines/pl-stock-prices-ut-stack.json",
                    "source": "./tmp-pl-stock-prices-ut-stack.json",
                    "depends_on": ["databricks_pipeline.pl-custom-name"],
                    "provider": "databricks",
                }
            },
        },
    }

    # Dev
    data = stack.to_terraform(env="dev").model_dump()
    data["provider"]["databricks"]["token"] = "***"
    data0 = copy.deepcopy(data_default)
    print(data0["resource"]["databricks_job"]["job-stock-prices-ut-stack"])
    cluster = data0["resource"]["databricks_job"]["job-stock-prices-ut-stack"][
        "job_cluster"
    ][0]["new_cluster"]
    cluster["node_type_id"] = "Standard_DS3_v2"
    cluster["spark_env_vars"]["LAKTORY_WORKSPACE_ENV"] = "dev"
    assert data == data0

    # Prod
    data = stack.to_terraform(env="prod").model_dump()
    data["provider"]["databricks"]["token"] = "***"
    data0 = copy.deepcopy(data_default)
    cluster = data0["resource"]["databricks_job"]["job-stock-prices-ut-stack"][
        "job_cluster"
    ][0]["new_cluster"]
    cluster["node_type_id"] = "Standard_DS4_v2"
    cluster["spark_env_vars"]["LAKTORY_WORKSPACE_ENV"] = "prod"
    data0["resource"]["databricks_pipeline"]["pl-custom-name"]["development"] = False
    assert data == data0


def test_terraform_plan():
    tstack = stack.to_terraform(env="dev")
    tstack.terraform.backend = (
        None  # TODO: Add credentials to git actions to use azure backend
    )
    tstack.init(flags=["-migrate-state"])
    tstack.plan()


def test_all_resources():
    from tests.test_catalog import catalog
    from tests.test_directory import directory
    from tests.test_job import job
    from tests.test_pipeline import pl
    from tests.test_metastore import metastore
    from tests.test_notebook import nb
    from tests.test_schema import schema
    from tests.test_sql_query import query
    from tests.test_user import user
    from tests.test_user import group
    from tests.test_workspacefile import workspace_file

    root_dir = os.path.dirname(__file__)

    nb.source = os.path.join(root_dir, nb.source)
    workspace_file.source = os.path.join(root_dir, workspace_file.source)

    validator = StackValidator(
        resources={
            "catalogs": [catalog],
            "directories": [directory],
            "jobs": [job],
            "metastores": [metastore],
            "pipelines": [pl],  # required by job
            "notebooks": [nb],
            "schemas": [schema],
            "sqlqueries": [query],
            "groups": [group],
            "users": [user],
            "workspacefiles": [workspace_file],
        },
        providers={
            "provider-workspace-neptune": {
                "host": "${vars.DATABRICKS_HOST}",
                # "azure_client_id": "0",
                # "azure_client_secret": "0",
                # "azure_tenant_id": "0",
            }
        },
    )

    validator.validate()


if __name__ == "__main__":
    test_stack_model()
    test_pulumi_stack()
    test_pulumi_preview()
    test_terraform_stack()
    test_terraform_plan()
    test_all_resources()



import pandas as pd
import numpy as np
import pytest
from uuid import UUID
from pyspark.sql import SparkSession
from pyspark.sql import functions as F

import laktory.spark.functions as LF

pdf = pd.DataFrame(
    {
        "x": [1, 2, 3],
        "a": [1, -1, 1],
        "b": [2, 0, 2],
        "c": [3, 0, 3],
        "n": [4, 0, 4],
        "pi": [np.pi] * 3,
        "p": [2, 0.2, 0.05],
        "word": ["dog_cat", "dog_cat_mouse", "dog"],
    },
)
spark = SparkSession.builder.appName("UnitTesting").getOrCreate()

df0 = spark.createDataFrame(pdf)


def test_compare(df0=df0):
    df = df0.withColumn("compare1", LF.compare("x", "a"))
    df = df.withColumn(
        "compare2", LF.compare("x", "a", operator=">", where=F.col("a") > 0)
    )
    pdf = df.toPandas()

    assert pdf["compare1"].tolist() == [True, False, False]
    assert pdf["compare2"].tolist() == [False, None, True]


def test_poly(df0=df0):
    df = df0.withColumn("poly1_1", LF.poly1("x", -1, 1.0))
    df = df.withColumn("poly1_2", LF.poly1("x", F.col("a"), F.col("b")))
    df = df.withColumn("poly2", LF.poly2("x", 1, c=-1))
    pdf = df.toPandas()

    assert pdf["poly1_1"].tolist() == [0, -1, -2]
    assert pdf["poly1_2"].tolist() == [3, -2, 5]
    assert pdf["poly1_2"].tolist() == [3, -2, 5]
    assert pdf["poly2"].tolist() == [0, 3, 8]


def test_power(df0=df0):
    df = df0.withColumn("power", LF.power("x", n=F.col("b")))
    pdf = df.toPandas()

    assert pdf["power"].tolist() == [1, 1, 9]


def test_roundp(df0=df0):
    df = df0.withColumn("roundp_1", LF.roundp("pi", p=0.2))
    df = df.withColumn("roundp_2", LF.roundp("pi", p=F.col("p")))
    pdf = df.toPandas()
    assert pdf["roundp_1"].tolist() == pytest.approx([3.2, 3.2, 3.2], abs=0.0001)
    assert pdf["roundp_2"].tolist() == pytest.approx([4, 3.20, 3.15], abs=0.0001)


def test_string_split(df0=df0):
    df = df0.withColumn("split_1", LF.string_split(F.col("word"), "_", 0))
    df = df.withColumn("split_2", LF.string_split(F.col("word"), "_", 1))
    pdf = df.toPandas()

    assert pdf["split_1"].tolist() == ["dog", "dog", "dog"]
    assert pdf["split_2"].tolist() == ["cat", "cat", None]


def test_uuid(df0=df0):
    df = df0.withColumn("uuid", LF.uuid())
    pdf = df.toPandas()

    for _uuid in pdf["uuid"]:
        assert str(UUID(_uuid)) == _uuid

    assert pdf["uuid"].nunique() == 3


def test_units(df0=df0):
    df = df0.withColumn("ft", LF.convert_units("x", "m", "ft"))
    df = df.withColumn("kelvin", LF.convert_units("x", "C", "K"))
    pdf = df.toPandas()

    assert pdf["ft"].tolist() == [
        3.280839895013124,
        6.561679790026248,
        9.842519685039372,
    ]
    assert pdf["kelvin"].tolist() == [274.15, 275.15, 276.15]


if __name__ == "__main__":
    test_compare()
    test_poly()
    test_power()
    test_roundp()
    test_string_split()
    test_uuid()
    test_units()



from abc import abstractmethod
from typing import Union
from typing import Any
from laktory._settings import settings
from laktory._parsers import _snake_to_camel
from laktory.models.resources.baseresource import BaseResource

# TODO: Review pulumi_outputs and pulumi_resources as they cause issues when
# running multiple tests in parallel

pulumi_outputs = {}
"""Pulumi outputs for all deployed resources. Updated during deployment."""

pulumi_resources = {}
"""All pulumi deployed resources objects. Updated during deployment."""


class PulumiResource(BaseResource):
    """
    Parent class for all Laktory models deployable with Pulumi IaC backend.
    """

    _pulumi_resources: dict[str, Any] = {}

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    @abstractmethod
    def pulumi_resource_type(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def pulumi_cls(self):
        raise NotImplementedError()

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        """List of fields to exclude when dumping model to pulumi"""
        return []

    @property
    def pulumi_renames(self) -> dict[str, str]:
        """Map of fields to rename when dumping model to pulumi"""
        return {}

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_properties(self) -> dict:
        """
        Resources properties formatted for pulumi:

        * Serialization (model dump)
        * Removal of excludes defined in `self.pulumi_excludes`
        * Renaming of keys according to `self.pulumi_renames`
        * Injection of variables

        Returns
        -------
        :
            Pulumi-safe model dump
        """
        d = super().model_dump(exclude=self.pulumi_excludes, exclude_none=True)
        for k, v in self.pulumi_renames.items():
            if settings.camel_serialization:
                k = _snake_to_camel(k)
                v = _snake_to_camel(v)
            if k in d:
                d[v] = d.pop(k)
        d = self.inject_vars(d)
        return d

    def to_pulumi(self, opts=None):
        """
        Instantiate one or multiple pulumi objects related to the model. This
        function would trigger a deployment in the context of a `pulumi up`

        Parameters
        ----------
        opts:
            Resource options for deployment

        Returns
        -------
        :
            Deployed pulumi resources
        """
        from pulumi import ResourceOptions

        self._pulumi_resources = {}

        for r in self.core_resources:
            # Properties
            properties = r.pulumi_properties
            properties = self.inject_vars(properties)

            # Options
            _opts = r.options.model_dump()
            if _opts is None:
                _opts = {}
            _opts = self.inject_vars(_opts)
            _opts = ResourceOptions(**_opts)
            if opts is not None:
                _opts = ResourceOptions.merge(_opts, opts)

            _r = r.pulumi_cls(r.resource_name, **properties, opts=_opts)

            # Save resource
            self._pulumi_resources[r.resource_name] = _r
            pulumi_resources[r.resource_name] = _r

            # Save resource outputs
            # TODO: Store other properties (like url, etc.).
            for k in [
                "id",
                "object_id",
                "path",
            ]:
                if hasattr(_r, k):
                    pulumi_outputs[f"{r.resource_name}.{k}"] = getattr(_r, k)

        # Return resources
        return self._pulumi_resources



from laktory.spark import DataFrame
from typing import Union
from typing import Literal
from typing import Any
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory._logger import get_logger

logger = get_logger(__name__)


class Watermark(BaseModel):
    """
    Definition of a spark structured streaming watermark for joining data
    streams.

    Attributes
    ----------
    column:
        Event time column name
    threshold:
        How late, expressed in seconds, the data is expected to be with
        respect to event time.

    References
    ----------
    https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#handling-late-data-and-watermarking
    """

    column: str
    threshold: str


class TableDataSourceCDC(BaseModel):
    """
    Defines the change data capture (CDC) properties of a table data source.
    They are used to build the target using `apply_changes` method from
    Databricks DLT.

    Attributes
    ----------
    apply_as_deletes:
        Specifies when a CDC event should be treated as a DELETE rather than
        an upsert. To handle out-of-order data, the deleted row is temporarily
        retained as a tombstone in the underlying Delta table, and a view is
        created in the metastore that filters out these tombstones.
    apply_as_truncates:
        Specifies when a CDC event should be treated as a full table TRUNCATE.
        Because this clause triggers a full truncate of the target table, it
        should be used only for specific use cases requiring this
        functionality.
    columns:
        A subset of columns to include in the target table. Use `columns` to
        specify the complete list of columns to include.
    except_columns:
        A subset of columns to exclude in the target table.
    ignore_null_updates:
        Allow ingesting updates containing a subset of the target columns.
        When a CDC event matches an existing row and ignore_null_updates is
        `True`, columns with a null will retain their existing values in the
        target. This also applies to nested columns with a value of null. When
        ignore_null_updates is `False`, existing values will be overwritten
        with null values.
    primary_keys:
        The column or combination of columns that uniquely identify a row in
        the source data. This is used to identify which CDC events apply to
        specific records in the target table.
    scd_type:
        Whether to store records as SCD type 1 or SCD type 2.
    sequence_by:
        The column name specifying the logical order of CDC events in the
        source data. Delta Live Tables uses this sequencing to handle change
        events that arrive out of order.
    track_history_columns:
        A subset of output columns to be tracked for history in the target table.
    track_history_except_columns:
        A subset of output columns to be excluded from tracking.

    References
    ----------
    https://docs.databricks.com/en/delta-live-tables/python-ref.html#change-data-capture-with-python-in-delta-live-tables
    """

    apply_as_deletes: Union[str, None] = None
    apply_as_truncates: Union[str, None] = None
    columns: Union[list[str], None] = []
    except_columns: Union[list[str], None] = []
    ignore_null_updates: Union[bool, None] = None
    primary_keys: list[str]
    scd_type: Literal[1, 2] = None
    sequence_by: str
    track_history_columns: Union[list[str], None] = None
    track_history_except_columns: Union[list[str], None] = None


class TableDataSource(BaseDataSource):
    """
    Data source using a SQL table, generally used in the context of a
    data pipeline.

    Attributes
    ----------
    catalog_name:
        Name of the catalog of the source table
    cdc:
        Change data capture specifications
    selects:
        Columns to select from the source table. Can be specified as a list
        or as a dictionary to rename the source columns
    filter:
        SQL expression used to select specific rows from the source table
    fmt:
        Table format
    from_pipeline:
        If `True` the source table will be read using `dlt.read` instead of
        `spark.read`
    name:
        Name of the source table
    path:
        Path of the source table
    schema_name:
        Name of the schema of the source table
    watermark
        Spark structured streaming watermark specifications

    Examples
    ---------
    ```python
    from laktory import models

    source = models.TableDataSource(
        name="brz_stock_prices",
        selects=["symbol", "open", "close"],
        filter="symbol='AAPL'",
        from_pipeline=False,
        read_as_stream=True,
    )
    # df = source.read(spark)
    ```
    """

    _df: Any = None
    catalog_name: Union[str, None] = None
    cdc: Union[TableDataSourceCDC, None] = None
    selects: Union[list[str], dict[str, str], None] = None
    fmt: Literal["PARQUET", "DELTA"] = "DELTA"
    filter: Union[str, None] = None
    from_pipeline: Union[bool, None] = True
    name: Union[str, None] = None
    path: Union[str, None] = None
    schema_name: Union[str, None] = None
    watermark: Union[Watermark, None] = None

    @model_validator(mode="after")
    def set_source(self) -> Any:
        """Source Definition"""

        if self.name and self.path:
            raise ValueError("Either path or name must be specified, but not both")

        if self.path:
            self.from_pipeline = False

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        if self.name is None:
            return None

        name = ""
        if self.catalog_name is not None:
            name = self.catalog_name

        if self.schema_name is not None:
            if name == "":
                name = self.schema_name
            else:
                name += f".{self.schema_name}"

        if name == "":
            name = self.name
        else:
            name += f".{self.name}"

        return name

    @property
    def from_path(self):
        return self.path is not None

    @property
    def path_or_full_name(self):
        if self.from_path:
            return self.path
        return self.full_name

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read(self, spark) -> DataFrame:
        from laktory.dlt import read
        from laktory.dlt import read_stream

        if self._df is not None:
            logger.info(f"Reading {self.path_or_full_name} from memory")
            df = self._df
        elif self.read_as_stream:
            logger.info(f"Reading {self.path_or_full_name} as stream")
            if self.from_pipeline:
                df = read_stream(self.full_name)
            elif self.from_path:
                df = spark.readStream.format(self.fmt).load(self.path)
            else:
                df = spark.readStream.format(self.fmt).table(self.full_name)
        else:
            logger.info(f"Reading {self.path_or_full_name} as static")
            if self.from_pipeline:
                df = read(self.full_name)
            elif self.from_path:
                df = spark.read.format(self.fmt).load(self.path)
            else:
                df = spark.read.format(self.fmt).table(self.full_name)

        return df

    def read(self, spark) -> DataFrame:
        import pyspark.sql.functions as F

        df = self._read(spark)

        # Apply filter
        if self.filter:
            df = df.filter(self.filter)

        # Columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [F.col(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [F.col(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Apply Watermark
        if self.watermark:
            df = df.withWatermark(
                self.watermark.column,
                self.watermark.threshold,
            )

        return df


if __name__ == "__main__":
    from laktory import models

    source = models.TableDataSource(
        name="brz_stock_prices",
        selects=["symbol", "open", "close"],
        filter="symbol='AAPL'",
        from_pipeline=False,
        read_as_stream=True,
    )
    df = source.read(spark)



from typing import Union
from typing import Any
from typing import Literal
from pydantic import model_validator

from laktory._logger import get_logger
from laktory._parsers import merge_dicts
from laktory.models.basemodel import BaseModel
from laktory.models.databricks.cluster import Cluster
from laktory.models.databricks.directory import Directory
from laktory.models.databricks.group import Group
from laktory.models.databricks.job import Job
from laktory.models.databricks.externallocation import ExternalLocation
from laktory.models.databricks.metastore import Metastore
from laktory.models.databricks.metastoredataaccess import MetastoreDataAccess
from laktory.models.databricks.notebook import Notebook
from laktory.models.databricks.pipeline import Pipeline
from laktory.models.databricks.secret import Secret
from laktory.models.databricks.secretscope import SecretScope
from laktory.models.databricks.serviceprincipal import ServicePrincipal
from laktory.models.databricks.sqlquery import SqlQuery
from laktory.models.databricks.user import User
from laktory.models.databricks.warehouse import Warehouse
from laktory.models.databricks.workspacefile import WorkspaceFile
from laktory.models.providers.baseprovider import BaseProvider
from laktory.models.providers.awsprovider import AWSProvider
from laktory.models.providers.azureprovider import AzureProvider
from laktory.models.providers.azurepulumiprovider import AzurePulumiProvider
from laktory.models.providers.databricksprovider import DatabricksProvider
from laktory.models.sql.catalog import Catalog
from laktory.models.sql.schema import Schema
from laktory.models.sql.table import Table
from laktory.models.sql.volume import Volume

logger = get_logger(__name__)

DIRPATH = "./"


class Terraform(BaseModel):
    backend: dict[str, Any] = None


class Pulumi(BaseModel):
    """
    config:
        Pulumi configuration settings. Generally used to
        configure providers. See references for more details.
    outputs:
        Requested resources-related outputs. See references for details.

    References
    ----------
    - Pulumi [configuration](https://www.pulumi.com/docs/concepts/config/)
    - Pulumi [outputs](https://www.pulumi.com/docs/concepts/inputs-outputs/#outputs)
    """

    config: dict[str, str] = {}
    outputs: dict[str, str] = {}


class StackResources(BaseModel):
    """
    Resources definition for a given stack or stack environment.

    Attributes
    ----------
    catalogs:
        Catalogs
    clusters:
        Clusters
    directories:
        Directories
    externallocations:
        External Locations
    groups:
        Groups
    jobs:
        Jobs
    metastores:
        Metastores
    notebooks:
        Notebooks
    pipelines:
        Pipelines
    schemas:
        Schemas
    secretscopes:
        SecretScopes
    serviceprincipals:
        ServicePrincipals
    sqlqueries:
        SQLQueries
    tables:
        Tables
    providers:
        Providers
    users:
        Users
    volumes:
        Volumes
    warehouses:
        Warehouses
    workspacefiles:
        WorkspacFiles
    """

    catalogs: dict[str, Catalog] = {}
    clusters: dict[str, Cluster] = {}
    directories: dict[str, Directory] = {}
    externallocations: dict[str, ExternalLocation] = {}
    groups: dict[str, Group] = {}
    jobs: dict[str, Job] = {}
    metastoredataaccesses: dict[str, MetastoreDataAccess] = {}
    metastores: dict[str, Metastore] = {}
    notebooks: dict[str, Notebook] = {}
    pipelines: dict[str, Pipeline] = {}
    schemas: dict[str, Schema] = {}
    secrets: dict[str, Secret] = {}
    secretscopes: dict[str, SecretScope] = {}
    serviceprincipals: dict[str, ServicePrincipal] = {}
    sqlqueries: dict[str, SqlQuery] = {}
    tables: dict[str, Table] = {}
    providers: dict[
        str, Union[AWSProvider, AzureProvider, AzurePulumiProvider, DatabricksProvider]
    ] = {}
    users: dict[str, User] = {}
    volumes: dict[str, Volume] = {}
    warehouses: dict[str, Warehouse] = {}
    workspacefiles: dict[str, WorkspaceFile] = {}

    @model_validator(mode="after")
    def update_resource_names(self) -> Any:
        for k, r in self._get_all().items():
            if r.resource_name_ and k != r.resource_name_:
                raise ValueError(
                    f"Provided resource name {r.resource_name_} does not match provided key {k}"
                )
            r.resource_name_ = k
        return self

    def _get_all(self, providers_excluded=False, providers_only=False):
        resources = {}
        for resource_type in self.model_fields.keys():
            if resource_type in ["variables"]:
                continue

            for resource_name, _r in getattr(self, resource_type).items():
                if providers_excluded and isinstance(_r, BaseProvider):
                    continue

                if providers_only and not isinstance(_r, BaseProvider):
                    continue

                resources[resource_name] = _r

        return resources


class EnvironmentStack(BaseModel):
    """
    Environment-specific stack definition.

    Attributes
    ----------
    backend:
        IaC backend used for deployment.
    description:
        Description of the stack
    name:
        Name of the stack. If Pulumi is used as a backend, it should match
        the name of the Pulumi project.
    organization:
        Organization
    pulumi:
        Pulumi-specific settings
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    terraform:
        Terraform-specific settings
    variables:
        Dictionary of variables made available in the resources definition.
    """

    backend: Literal["pulumi", "terraform"] = None
    description: str = None
    name: str
    organization: str = None
    pulumi: Pulumi = Pulumi()
    resources: StackResources = StackResources()
    terraform: Terraform = Terraform()
    variables: dict[str, Any] = {}


class EnvironmentSettings(BaseModel):
    """
    Settings overwrite for a specific environments

    Attributes
    ----------
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    variables:
        Dictionary of variables made available in the resources definition.
    """

    resources: Any = None
    variables: dict[str, Any] = None


class Stack(BaseModel):
    """
    The Stack defines a collection of deployable resources, the deployment
    configuration, some variables and the environment-specific settings.

    Attributes
    ----------
    backend:
        IaC backend used for deployment.
    description:
        Description of the stack
    environments:
        Environment-specific overwrite of config, resources or variables
        arguments.
    name:
        Name of the stack. If Pulumi is used as a backend, it should match
        the name of the Pulumi project.
    organization:
        Organization
    pulumi:
        Pulumi-specific settings
    resources:
        Dictionary of resources to be deployed. Each key should be a resource
        type and each value should be a dictionary of resources who's keys are
        the resource names and the values the resources definitions.
    terraform:
        Terraform-specific settings
    variables:
        Dictionary of variables made available in the resources definition.

    Examples
    --------
    ```py
    from laktory import models

    stack = models.Stack(
        name="workspace",
        backend="pulumi",
        pulumi={
            "config": {
                "databricks:host": "${vars.DATABRICKS_HOST}",
                "databricks:token": "${vars.DATABRICKS_TOKEN}",
            },
        },
        resources={
            "pipelines": {
                "pl-stock-prices": {
                    "name": "pl-stock-prices",
                    "development": "${vars.is_dev}",
                    "libraries": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}},
                    ],
                }
            },
            "jobs": {
                "job-stock-prices": {
                    "name": "job-stock-prices",
                    "clusters": [
                        {
                            "name": "main",
                            "spark_version": "14.0.x-scala2.12",
                            "node_type_id": "Standard_DS3_v2",
                        }
                    ],
                    "tasks": [
                        {
                            "task_key": "ingest",
                            "job_cluster_key": "main",
                            "notebook_task": {
                                "notebook_path": "/.laktory/jobs/ingest_stock_prices.py",
                            },
                        },
                        {
                            "task_key": "pipeline",
                            "depends_ons": [{"task_key": "ingest"}],
                            "pipeline_task": {
                                "pipeline_id": "${resources.pl-stock-prices.id}",
                            },
                        },
                    ],
                }
            },
        },
        variables={
            "org": "okube",
        },
        environments={
            "dev": {
                "variables": {
                    "is_dev": True,
                }
            },
            "prod": {
                "variables": {
                    "is_dev": False,
                }
            },
        },
    )

    print(stack)
    '''
    variables={'org': 'okube'} backend='pulumi' description=None environments={'dev': EnvironmentSettings(variables={'is_dev': True}, resources=None), 'prod': EnvironmentSettings(variables={'is_dev': False}, resources=None)} name='workspace' organization=None pulumi=Pulumi(variables={}, config={'databricks:host': '${vars.DATABRICKS_HOST}', 'databricks:token': '${vars.DATABRICKS_TOKEN}'}, outputs={}) resources=StackResources(variables={}, catalogs={}, clusters={}, directories={}, externallocations={}, groups={}, jobs={'job-stock-prices': Job(resource_name_='job-stock-prices', options=ResourceOptions(variables={}, depends_on=[], provider=None, aliases=None, delete_before_replace=True, ignore_changes=None, import_=None, parent=None, replace_on_changes=None), variables={}, access_controls=[], clusters=[JobCluster(resource_name_=None, options=ResourceOptions(variables={}, depends_on=[], provider=None, aliases=None, delete_before_replace=True, ignore_changes=None, import_=None, parent=None, replace_on_changes=None), variables={}, access_controls=None, apply_policy_default_values=None, autoscale=None, autotermination_minutes=None, cluster_id=None, custom_tags=None, data_security_mode='USER_ISOLATION', driver_instance_pool_id=None, driver_node_type_id=None, enable_elastic_disk=None, enable_local_disk_encryption=None, idempotency_token=None, init_scripts=[], instance_pool_id=None, is_pinned=None, libraries=None, name='main', node_type_id='Standard_DS3_v2', num_workers=None, policy_id=None, runtime_engine=None, single_user_name=None, spark_conf={}, spark_env_vars={}, spark_version='14.0.x-scala2.12', ssh_public_keys=[])], continuous=None, control_run_state=None, email_notifications=None, format=None, health=None, max_concurrent_runs=None, max_retries=None, min_retry_interval_millis=None, name='job-stock-prices', notification_settings=None, parameters=[], retry_on_timeout=None, run_as=None, schedule=None, tags={}, tasks=[JobTask(variables={}, condition_task=None, depends_ons=None, description=None, email_notifications=None, existing_cluster_id=None, health=None, job_cluster_key='main', libraries=None, max_retries=None, min_retry_interval_millis=None, notebook_task=JobTaskNotebookTask(variables={}, notebook_path='/.laktory/jobs/ingest_stock_prices.py', base_parameters=None, source=None), notification_settings=None, pipeline_task=None, retry_on_timeout=None, run_if=None, run_job_task=None, sql_task=None, task_key='ingest', timeout_seconds=None), JobTask(variables={}, condition_task=None, depends_ons=[JobTaskDependsOn(variables={}, task_key='ingest', outcome=None)], description=None, email_notifications=None, existing_cluster_id=None, health=None, job_cluster_key=None, libraries=None, max_retries=None, min_retry_interval_millis=None, notebook_task=None, notification_settings=None, pipeline_task=JobTaskPipelineTask(variables={}, pipeline_id='${resources.pl-stock-prices.id}', full_refresh=None), retry_on_timeout=None, run_if=None, run_job_task=None, sql_task=None, task_key='pipeline', timeout_seconds=None)], timeout_seconds=None, trigger=None, webhook_notifications=None)}, metastoredataaccesses={}, metastores={}, notebooks={}, pipelines={'pl-stock-prices': Pipeline(resource_name_='pl-stock-prices', options=ResourceOptions(variables={}, depends_on=[], provider=None, aliases=None, delete_before_replace=True, ignore_changes=None, import_=None, parent=None, replace_on_changes=None), variables={}, access_controls=[], allow_duplicate_names=None, catalog=None, channel='PREVIEW', clusters=[], configuration={}, continuous=None, development='${vars.is_dev}', edition=None, libraries=[PipelineLibrary(variables={}, file=None, notebook=PipelineLibraryNotebook(variables={}, path='/pipelines/dlt_brz_template.py'))], name='pl-stock-prices', notifications=[], photon=None, serverless=None, storage=None, tables=[], target=None, udfs=[])}, schemas={}, secrets={}, secretscopes={}, serviceprincipals={}, sqlqueries={}, tables={}, providers={}, users={}, volumes={}, warehouses={}, workspacefiles={}) terraform=Terraform(variables={}, backend=None)
    '''
    ```

    """

    backend: Literal["pulumi", "terraform"] = None
    description: str = None
    environments: dict[str, EnvironmentSettings] = {}
    name: str
    organization: str = None
    pulumi: Pulumi = Pulumi()
    resources: StackResources = StackResources()
    terraform: Terraform = Terraform()
    variables: dict[str, Any] = {}

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def envs(self) -> dict[str, EnvironmentStack]:
        """
        Complete definition of each of the stack environments. It takes into
        account both the default stack values and environment-specific
        overwrites.

        Returns
        -------
        :
            Environment definitions.
        """
        ENV_FIELDS = ["pulumi", "resources", "terraform", "variables"]

        d = self.model_dump(exclude_none=True)
        _envs = d.pop("environments")

        # Restore fields excluded from dump
        for rtype in d["resources"]:
            for k, r in d["resources"][rtype].items():
                r["options"] = getattr(self.resources, rtype)[k].options.model_dump(
                    exclude_none=True
                )

        envs = {}
        for env_name, env in self.environments.items():
            for k in ENV_FIELDS:
                d[k] = merge_dicts(d[k], _envs[env_name].get(k, {}))

            envs[env_name] = EnvironmentStack(**d)

        return envs

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    def to_pulumi(self, env: Union[str, None] = None):
        """
        Create a pulumi stack for a given environment `env`.

        Parameters
        ----------
        env:
            Target environment. If `None`, used default stack values only.

        Returns
        -------
        : PulumiStack
            Pulumi-specific stack definition
        """
        from laktory.models.stacks.pulumistack import PulumiStack

        if env is not None and env in self.envs:
            env = self.envs[env]
        else:
            env = self

        # Resources
        resources = {}
        for r in env.resources._get_all().values():
            r.variables = env.variables
            for _r in r.core_resources:
                resources[_r.resource_name] = _r

        return PulumiStack(
            name=env.name,
            organization=env.organization,
            config=env.pulumi.config,
            description=env.description,
            resources=resources,
            variables=env.variables,
            outputs=env.pulumi.outputs,
        )

    # ----------------------------------------------------------------------- #
    # Terraform Methods                                                       #
    # ----------------------------------------------------------------------- #

    def to_terraform(self, env: Union[str, None] = None):
        """
        Create a terraform stack for a given environment `env`.

        Parameters
        ----------
        env:
            Target environment. If `None`, used default stack values only.

        Returns
        -------
        : TerraformStack
            Terraform-specific stack definition
        """
        from laktory.models.stacks.terraformstack import TerraformStack

        if env is not None and env in self.envs:
            env = self.envs[env]
        else:
            env = self

        # Providers
        providers = {}
        for r in env.resources._get_all(providers_only=True).values():
            for _r in r.core_resources:
                rname = _r.resource_name
                providers[rname] = _r

        # Resources
        resources = {}
        for r in env.resources._get_all(providers_excluded=True).values():
            r.variables = env.variables
            for _r in r.core_resources:
                resources[_r.resource_name] = _r

        # Update terraform
        return TerraformStack(
            terraform={"backend": env.terraform.backend},
            providers=providers,
            resources=resources,
            variables=env.variables,
        )



from pydantic import model_validator
from laktory.spark import DataFrame
from typing import Literal
from typing import Any

from laktory.models.dataeventheader import DataEventHeader
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory._logger import get_logger

logger = get_logger(__name__)


class EventDataSource(BaseDataSource, DataEventHeader):
    """
    Data source using events data (files), generally used in the context of a
    data pipeline.

    Attributes
    ----------
    fmt
        Format of the stored data
    header
        If `True`, first line of CSV files is assumed to be the column names.
    multiline
        If `True`, JSON files are parsed assuming that an object maybe be
        defined on multiple lines (as opposed to having a single object
        per line)
    read_options:
        Other options passed to `spark.read.options`
    schema_location:
        Path for events schema. If `None`, `event_root` is used.
    type
        Type of infrastructure storing the data

    Examples
    ---------
    ```python
    from laktory import models

    source = models.EventDataSource(
        name="stock_price",
        producer={"name": "yahoo-finance"},
        fmt="JSON",
        read_as_stream=False,
    )
    # df = source.read(spark)
    ```
    """

    fmt: Literal["JSON", "CSV", "PARQUET", "DELTA"] = "JSON"
    header: bool = True
    multiline: bool = False
    read_options: dict[str, str] = {}
    schema_location: str = None
    type: Literal["STORAGE_EVENTS", "STREAM_KINESIS", "STREAM_KAFKA"] = "STORAGE_EVENTS"

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_storage(self, spark) -> DataFrame:
        if self.read_as_stream:
            logger.info(f"Reading {self.event_root} as stream")

            schema_location = self.schema_location
            if schema_location is None:
                schema_location = self.event_root

            # Set reader
            reader = (
                spark.readStream.format("cloudFiles")
                .option("cloudFiles.format", self.fmt)
                .option("cloudFiles.schemaLocation", schema_location)
                .option("cloudFiles.inferColumnTypes", True)
                .option("cloudFiles.schemaEvolutionMode", "addNewColumns")
                .option("cloudFiles.allowOverwrites", True)
            )

        else:
            logger.info(f"Reading {self.event_root} as static")

            # Set reader
            reader = spark.read.format(self.fmt)

        reader = (
            reader.option("multiLine", self.multiline)  # only apply to JSON format
            .option("mergeSchema", True)
            .option("recursiveFileLookup", True)
            .option("header", self.header)  # only apply to CSV format
        )
        if self.read_options:
            reader = reader.options(**self.read_options)

        # Load
        df = reader.load(self.event_root)

        # Not supported by UC
        # .withColumn("file", F.input_file_name())

        return df

    def read(self, spark) -> DataFrame:
        """
        Read data with options specified in attributes.

        Parameters
        ----------
        spark
            Spark context

        Returns
        -------
        : DataFrame
            Resulting park dataframe
        """
        if self.type == "STORAGE_EVENTS":
            return self._read_storage(spark)
        else:
            raise NotImplementedError()



from laktory._testing import StockPricesPipeline

pl = StockPricesPipeline()


def test_pipeline():
    print(pl.model_dump())
    assert pl.model_dump() == {
        "access_controls": [],
        "allow_duplicate_names": None,
        "catalog": None,
        "channel": "PREVIEW",
        "clusters": [],
        "configuration": {},
        "continuous": None,
        "development": None,
        "edition": None,
        "libraries": None,
        "name": "pl-stock-prices",
        "notifications": [],
        "photon": None,
        "serverless": None,
        "storage": None,
        "tables": [
            {
                "builder": {
                    "aggregation": None,
                    "as_dlt_view": False,
                    "drop_columns": [],
                    "drop_duplicates": None,
                    "drop_source_columns": False,
                    "event_source": {
                        "name": "stock_price",
                        "description": None,
                        "producer": None,
                        "events_root_": None,
                        "event_root_": None,
                        "read_as_stream": True,
                        "fmt": "JSON",
                        "header": True,
                        "multiline": False,
                        "read_options": {},
                        "schema_location": None,
                        "type": "STORAGE_EVENTS",
                    },
                    "filter": None,
                    "joins": [],
                    "joins_post_aggregation": [],
                    "layer": "BRONZE",
                    "pipeline_name": "pl-stock-prices",
                    "selects": None,
                    "table_source": None,
                    "template": "BRONZE",
                    "window_filter": None,
                },
                "catalog_name": None,
                "columns": [],
                "comment": None,
                "data": None,
                "data_source_format": "DELTA",
                "expectations": [],
                "grants": None,
                "name": "brz_stock_prices",
                "primary_key": None,
                "schema_name": None,
                "table_type": "MANAGED",
                "timestamp_key": None,
                "view_definition": None,
                "warehouse_id": "08b717ce051a0261",
            },
            {
                "builder": {
                    "aggregation": None,
                    "as_dlt_view": False,
                    "drop_columns": [],
                    "drop_duplicates": None,
                    "drop_source_columns": True,
                    "event_source": None,
                    "filter": None,
                    "joins": [],
                    "joins_post_aggregation": [],
                    "layer": "SILVER",
                    "pipeline_name": "pl-stock-prices",
                    "selects": None,
                    "table_source": {
                        "read_as_stream": True,
                        "catalog_name": "dev",
                        "cdc": None,
                        "selects": None,
                        "fmt": "DELTA",
                        "filter": None,
                        "from_pipeline": True,
                        "name": "brz_stock_prices",
                        "path": None,
                        "schema_name": "markets",
                        "watermark": None,
                    },
                    "template": "SILVER",
                    "window_filter": None,
                },
                "catalog_name": None,
                "columns": [
                    {
                        "catalog_name": None,
                        "comment": None,
                        "name": "created_at",
                        "pii": None,
                        "schema_name": None,
                        "spark_func_args": [
                            {
                                "value": "data._created_at",
                                "is_column": True,
                                "to_lit": False,
                                "to_expr": True,
                            }
                        ],
                        "spark_func_kwargs": {},
                        "spark_func_name": "coalesce",
                        "sql_expression": None,
                        "table_name": "slv_stock_prices",
                        "type": "timestamp",
                        "unit": None,
                    },
                    {
                        "catalog_name": None,
                        "comment": None,
                        "name": "symbol",
                        "pii": None,
                        "schema_name": None,
                        "spark_func_args": [
                            {
                                "value": "data.symbol",
                                "is_column": True,
                                "to_lit": False,
                                "to_expr": True,
                            }
                        ],
                        "spark_func_kwargs": {},
                        "spark_func_name": "coalesce",
                        "sql_expression": None,
                        "table_name": "slv_stock_prices",
                        "type": "string",
                        "unit": None,
                    },
                    {
                        "catalog_name": None,
                        "comment": None,
                        "name": "open",
                        "pii": None,
                        "schema_name": None,
                        "spark_func_args": [
                            {
                                "value": "data.open",
                                "is_column": True,
                                "to_lit": False,
                                "to_expr": True,
                            }
                        ],
                        "spark_func_kwargs": {},
                        "spark_func_name": "coalesce",
                        "sql_expression": None,
                        "table_name": "slv_stock_prices",
                        "type": "double",
                        "unit": None,
                    },
                    {
                        "catalog_name": None,
                        "comment": None,
                        "name": "close",
                        "pii": None,
                        "schema_name": None,
                        "spark_func_args": [],
                        "spark_func_kwargs": {},
                        "spark_func_name": None,
                        "sql_expression": "data.open",
                        "table_name": "slv_stock_prices",
                        "type": "double",
                        "unit": None,
                    },
                ],
                "comment": None,
                "data": [
                    ["2023-11-01T00:00:00Z", "AAPL", 1, 2],
                    ["2023-11-01T01:00:00Z", "AAPL", 3, 4],
                    ["2023-11-01T00:00:00Z", "GOOGL", 3, 4],
                    ["2023-11-01T01:00:00Z", "GOOGL", 5, 6],
                ],
                "data_source_format": "DELTA",
                "expectations": [
                    {
                        "name": "positive_price",
                        "expression": "open > 0",
                        "action": "FAIL",
                    },
                    {
                        "name": "recent_price",
                        "expression": "created_at > '2023-01-01'",
                        "action": "DROP",
                    },
                ],
                "grants": None,
                "name": "slv_stock_prices",
                "primary_key": None,
                "schema_name": None,
                "table_type": "MANAGED",
                "timestamp_key": None,
                "view_definition": None,
                "warehouse_id": "08b717ce051a0261",
            },
        ],
        "target": None,
        "udfs": [{"module_name": "stock_functions", "function_name": "high"}],
    }


def test_pipeline_pulumi():
    assert pl.resource_name == "pl-stock-prices"
    assert pl.options.model_dump(exclude_none=True) == {
        "depends_on": [],
        "delete_before_replace": True,
    }
    print(pl.pulumi_properties)
    assert pl.pulumi_properties == {
        "channel": "PREVIEW",
        "clusters": [],
        "configuration": {},
        "name": "pl-stock-prices",
        "notifications": [],
    }

    # Resources
    assert len(pl.core_resources) == 3
    r = pl.core_resources[-1]
    r.options.aliases = ["my-file"]
    assert pl.core_resources[-1].options.aliases == ["my-file"]


if __name__ == "__main__":
    test_pipeline()
    test_pipeline_pulumi()



SUPPORTED_TYPES = [
    "_any",
    "data",
    "void",
    "atomic",
    "numeric",
    "integral",
    "fractional",
    "string",
    "binary",
    "boolean",
    "date",
    "timestamp",
    "timestamp_ntz",
    "decimal(10,0)",
    "double",
    "float",
    "tinyint",
    "int",
    "bigint",
    "interval day to second",
    "smallint",
    "struct<>",
    "udt",
]

# CACHE_ROOT = "./laktory"
CACHE_ROOT = "./"
LAKTORY_WORKSPACE_ROOT = "/.laktory/"
# for k in vars(types):
#     o = getattr(types, k)
#     if "Type" in k and inspect.isclass(o):
#         try:
#             typ = o()
#         except TypeError:
#             continue
#
#         if not hasattr(typ, "simpleString"):
#             continue
#
#         SUPPORTED_TYPES[typ.simpleString()] = typ



import typer
from typing import Union
from typing_extensions import Annotated
from laktory.models.stacks.stack import Stack
from laktory.models.basemodel import BaseModel

APP_NAME = "laktory-cli"

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    help="Laktory CLI to preview and deploy resources",
)  # prevent display secret data


class CLIController(BaseModel):
    stack_filepath: Union[str, None] = None
    backend: Union[str, None] = None
    organization: Union[str, None] = None
    env: Union[str, None] = None
    pulumi_options_str: Union[str, None] = None
    terraform_options_str: Union[str, None] = None
    stack: Union[Stack, None] = None

    def model_post_init(self, __context):
        super().model_post_init(__context)

        # Read stack
        if self.stack_filepath is None:
            self.stack_filepath = "./stack.yaml"
        with open(self.stack_filepath, "r") as fp:
            self.stack = Stack.model_validate_yaml(fp)

        # Set backend
        if self.backend is None:
            self.backend = self.stack.backend
        if self.backend is None:
            raise ValueError(
                "backend ['pulumi', 'terraform'] must be specified in stack file or as CLI option "
            )

        # Set organization
        if self.organization is None:
            self.organization = self.stack.organization
        if self.organization is None and self.backend == "pulumi":
            raise ValueError("organization must be specified with pulumi backend")

        # Check environment
        if self.env is None and self.backend == "pulumi":
            raise ValueError("Environment must be specified with pulumi backend")

    @property
    def pulumi_options(self):
        if self.pulumi_options_str is None:
            return None
        return self.pulumi_options_str.split(",")

    @property
    def terraform_options(self):
        if self.terraform_options_str is None:
            return None
        return self.terraform_options_str.split(",")

    @property
    def pulumi_stack_name(self):
        return self.organization + "/" + self.env

    def pulumi_call(self, cmd):
        if self.pulumi_stack_name is None:
            raise ValueError("Argument `stack` must be specified with pulumi backend")

        pstack = self.stack.to_pulumi(env=self.env)
        getattr(pstack, cmd)(stack=self.pulumi_stack_name, flags=self.pulumi_options)

    def terraform_call(self, cmd):
        pstack = self.stack.to_terraform(env=self.env)
        getattr(pstack, cmd)(flags=self.terraform_options)


@app.command()
def write(
    backend: Annotated[str, typer.Option(help="IaC backend [terraform]")] = None,
    organization: Annotated[
        str,
        typer.Option(
            "--org",
            "-o",
            help="Name of the organization in associated with the pulumi stack.",
        ),
    ] = None,
    environment: Annotated[
        str, typer.Option("--env", "-e", help="Name of the environment")
    ] = None,
    filepath: Annotated[
        str, typer.Option(help="Stack (yaml) filepath.")
    ] = "./stack.yaml",
    pulumi_options: Annotated[
        str,
        typer.Option(
            "--pulumi-options", help="Comma separated pulumi options (flags)."
        ),
    ] = None,
    terraform_options: Annotated[
        str,
        typer.Option(
            "--terraform-options", help="Comma separated terraform options (flags)."
        ),
    ] = None,
):
    """
    Write IaC backend file

    Parameters
    ----------
    backend:
        IaC backend [terraform]
    organization:
        Name of the organization associated with the Pulumi stack.
    environment:
        Name of the environment.
    filepath:
        Stack (yaml) filepath.
    pulumi_options:
        Comma separated pulumi options (flags).
    terraform_options:
        Comma separated terraform options (flags).

    Examples
    --------
    ```cmd
    laktory write --env dev
    ```
    """
    controller = CLIController(
        backend=backend,
        organization=organization,
        env=environment,
        stack_filepath=filepath,
        pulumi_options_str=pulumi_options,
        terraform_options_str=terraform_options,
    )

    # Call
    if controller.backend == "pulumi":
        raise ValueError("Pulumi backend not supported for init command")
    elif controller.backend == "terraform":
        controller.terraform_call("init")
    else:
        raise ValueError("backend should be ['terraform']")


@app.command()
def init(
    backend: Annotated[str, typer.Option(help="IaC backend [terraform]")] = None,
    organization: Annotated[
        str,
        typer.Option(
            "--org",
            "-o",
            help="Name of the organization in associated with the pulumi stack.",
        ),
    ] = None,
    environment: Annotated[
        str, typer.Option("--env", "-e", help="Name of the environment")
    ] = None,
    filepath: Annotated[
        str, typer.Option(help="Stack (yaml) filepath.")
    ] = "./stack.yaml",
    pulumi_options: Annotated[
        str,
        typer.Option(
            "--pulumi-options", help="Comma separated pulumi options (flags)."
        ),
    ] = None,
    terraform_options: Annotated[
        str,
        typer.Option(
            "--terraform-options", help="Comma separated terraform options (flags)."
        ),
    ] = None,
):
    """
    Initialize IaC backend

    Parameters
    ----------
    backend:
        IaC backend [terraform]
    organization:
        Name of the organization associated with the Pulumi stack.
    environment:
        Name of the environment.
    filepath:
        Stack (yaml) filepath.
    pulumi_options:
        Comma separated pulumi options (flags).
    terraform_options:
        Comma separated terraform options (flags).

    Examples
    --------
    ```cmd
    laktory init --env dev
    ```

    References
    ----------
    - terraform [init](https://developer.hashicorp.com/terraform/cli/commands/init)
    """
    controller = CLIController(
        backend=backend,
        organization=organization,
        env=environment,
        stack_filepath=filepath,
        pulumi_options_str=pulumi_options,
        terraform_options_str=terraform_options,
    )

    # Call
    if controller.backend == "pulumi":
        raise ValueError("Pulumi backend not supported for init command")
    elif controller.backend == "terraform":
        controller.terraform_call("init")
    else:
        raise ValueError("backend should be ['terraform']")


@app.command()
def preview(
    backend: Annotated[
        str, typer.Option(help="IaC backend [pulumi, terraform]")
    ] = None,
    organization: Annotated[
        str,
        typer.Option(
            "--org",
            "-o",
            help="Name of the organization in associated with the pulumi stack.",
        ),
    ] = None,
    environment: Annotated[
        str, typer.Option("--env", "-e", help="Name of the environment")
    ] = None,
    filepath: Annotated[
        str, typer.Option(help="Stack (yaml) filepath.")
    ] = "./stack.yaml",
    pulumi_options: Annotated[
        str,
        typer.Option(
            "--pulumi-options", help="Comma separated pulumi options (flags)."
        ),
    ] = None,
    terraform_options: Annotated[
        str,
        typer.Option(
            "--terraform-options", help="Comma separated terraform options (flags)."
        ),
    ] = None,
):
    """
    Validate configuration and resources and preview deployment.

    Parameters
    ----------
    backend:
        IaC backend [pulumi, terraform]
    organization:
        Name of the organization associated with the Pulumi stack.
    environment:
        Name of the environment.
    filepath:
        Stack (yaml) filepath.
    pulumi_options:
        Comma separated pulumi options (flags).
    terraform_options:
        Comma separated terraform options (flags).

    Examples
    --------
    ```cmd
    laktory preview --env dev pulumi_options "--show-reads,--show-config"
    ```

    References
    ----------
    - pulumi preview [options](https://www.pulumi.com/docs/cli/commands/pulumi_preview/)
    """
    controller = CLIController(
        backend=backend,
        organization=organization,
        env=environment,
        stack_filepath=filepath,
        pulumi_options_str=pulumi_options,
        terraform_options_str=terraform_options,
    )

    # Call
    if controller.backend == "pulumi":
        controller.pulumi_call("preview")
    elif controller.backend == "terraform":
        controller.terraform_call("plan")
    else:
        raise ValueError("backend should be ['terraform', 'pulumi']")


@app.command()
def deploy(
    backend: Annotated[
        str, typer.Option(help="IaC backend [pulumi, terraform]")
    ] = None,
    organization: Annotated[
        str,
        typer.Option(
            "--org",
            "-o",
            help="Name of the organization in associated with the pulumi stack.",
        ),
    ] = None,
    environment: Annotated[
        str, typer.Option("--env", "-e", help="Name of the environment")
    ] = None,
    filepath: Annotated[
        str, typer.Option(help="Stack (yaml) filepath.")
    ] = "./stack.yaml",
    pulumi_options: Annotated[
        str,
        typer.Option(
            "--pulumi-options", help="Comma separated pulumi options (flags)."
        ),
    ] = None,
    terraform_options: Annotated[
        str,
        typer.Option(
            "--terraform-options", help="Comma separated terraform options (flags)."
        ),
    ] = None,
):
    """
    Execute deployment.

    Parameters
    ----------
    backend:
        IaC backend [pulumi, terraform]
    organization:
        Name of the organization associated with the Pulumi stack.
    environment:
        Name of the environment.
    filepath:
        Stack (yaml) filepath.
    pulumi_options:
        Comma separated pulumi options (flags).
    terraform_options:
        Comma separated terraform options (flags).

    Examples
    --------
    ```cmd
    laktory deploy --env dev --filepath my-stack.yaml
    ```

    References
    ----------
    - pulumi up [options](https://www.pulumi.com/docs/cli/commands/pulumi_up/)
    """
    controller = CLIController(
        backend=backend,
        organization=organization,
        env=environment,
        stack_filepath=filepath,
        pulumi_options_str=pulumi_options,
        terraform_options_str=terraform_options,
    )

    # Call
    if controller.backend == "pulumi":
        controller.pulumi_call("up")
    elif controller.backend == "terraform":
        controller.terraform_call("apply")
    else:
        raise ValueError("backend should be ['terraform', 'pulumi']")


if __name__ == "__main__":
    app()



from pydantic import field_validator
from pydantic import model_validator
from typing import Any
from typing import Callable
from typing import Union

from laktory._logger import get_logger
from laktory.constants import SUPPORTED_TYPES
from laktory.models.basemodel import BaseModel
from laktory.spark import Column as SparkColumn
from laktory.spark import DataFrame

logger = get_logger(__name__)


class SparkFuncArg(BaseModel):
    """
    Spark function argument

    Attributes
    ----------
    value:
        Value of the argument
    is_column:
        If `True`, the value is also the name of a column
    to_lit:
        If `True`, the value is converted to a spark column using
        `F.lit({value})`
    to_expr:
        If `True`, the value is converted to a spark column using
        `F.expr({value})`
    """

    value: Any
    is_column: bool = None
    to_lit: bool = None
    to_expr: bool = None

    @model_validator(mode="after")
    def set_defaults(self) -> Any:
        """Default options"""

        # If only value is provided and value is a string, assume value is column
        if self.is_column is None and self.to_expr is None and self.to_lit is None:
            if isinstance(self.value, str):
                self.is_column = True

        if self.is_column:
            if self.to_lit:
                raise ValueError(
                    "If `to_column` is `True`, `to_lit` must be `None` or `False`"
                )

            if self.to_expr == False:
                raise ValueError(
                    "If `to_column` is `True`, `to_expr` must be `None` or `True`"
                )

            self.to_expr = True
            self.to_lit = False

        else:
            self.is_column = False

            if self.to_lit is None and self.to_expr is None:
                self.to_lit = True
                self.to_expr = False

            if self.to_lit is None:
                self.to_lit = False

            if self.to_expr is None:
                self.to_expr = False

        return self

    def to_spark(self):
        import pyspark.sql.functions as F

        v = self.value
        if self.to_lit:
            v = F.lit(v)
        elif self.to_expr:
            v = F.expr(v)
        return v


class Column(BaseModel):
    """
    Definition of a table column, including instructions on how to build the
    column using spark and an input dataframe.

    Attributes
    ----------
    catalog_name:
        Name of the catalog string the column table
    comment:
        Text description of the column
    name:
        Name of the column
    pii:
        If `True`, the column is flagged as Personally Identifiable Information
    schema_name:
        Name of the schema storing the column table
    spark_func_args:
        List of arguments to be passed to the spark function to build the
        column.
    spark_func_kwargs:
        List of keyword arguments to be passed to the spark function to build
        the column.
    spark_func_name:
        Name of the spark function to build the column. Mutually exclusive to
        `sql_expression`
    sql_expression:
        Expression defining how to build the column. Mutually exclusive to
        `spark_func_name`
    table_name:
        Name of the table storing the column.
    type:
        Column data type
    unit:
        Column units

    Examples
    --------
    ```py
    from laktory import models

    df = spark.createDataFrame(
        [
            [200.0],
            [202.0],
            [201.5],
        ],
        ["data.open"],
    )

    col = models.Column(
        name="open",
        spark_func_name="poly1",
        spark_func_args=[
            "data.open",
        ],
        spark_func_kwargs={"a": {"value": 2.0, "to_lit": True}},
        type="double",
    )

    spark_col = col.to_spark(df)
    ```
    """

    catalog_name: Union[str, None] = None
    comment: Union[str, None] = None
    name: str
    pii: Union[bool, None] = None
    schema_name: Union[str, None] = None
    spark_func_args: list[Union[str, SparkFuncArg]] = []
    spark_func_kwargs: dict[str, Union[Any, SparkFuncArg]] = {}
    spark_func_name: Union[str, None] = None
    sql_expression: Union[str, None] = None
    table_name: Union[str, None] = None
    type: str = "string"
    unit: Union[str, None] = None

    @field_validator("spark_func_args")
    def parse_args(cls, args: list[Union[str, SparkFuncArg]]) -> list[SparkFuncArg]:
        _args = []
        for a in args:
            if isinstance(a, SparkFuncArg):
                pass
            elif isinstance(a, dict):
                a = SparkFuncArg(**a)
            else:
                a = SparkFuncArg(value=a)
            _args += [a]
        return _args

    @field_validator("spark_func_kwargs")
    def parse_kwargs(
        cls, kwargs: dict[str, Union[str, SparkFuncArg]]
    ) -> dict[str, SparkFuncArg]:
        _kwargs = {}
        for k, a in kwargs.items():
            if isinstance(a, SparkFuncArg):
                pass
            elif isinstance(a, dict):
                a = SparkFuncArg(**a)
            else:
                a = SparkFuncArg(value=a)
            _kwargs[k] = a
        return _kwargs

    @field_validator("type")
    def default_load_path(cls, v: str) -> str:
        if "<" in v:
            return v
        else:
            if v not in SUPPORTED_TYPES:
                raise ValueError(
                    f"Type {v} is not supported. Select one of {SUPPORTED_TYPES}"
                )
        return v

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Table full name `{catalog_name}.{schema_name}.{table_name}`"""
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        if self.table_name:
            if _id == "":
                _id = self.table_name
            else:
                _id += f".{self.table_name}"

        return _id

    @property
    def full_name(self) -> str:
        """Column full name `{catalog_name}.{schema_name}.{table_name}.{column_name}`"""
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    def to_spark(
        self,
        df: DataFrame,
        udfs: list[Callable[[...], SparkColumn]] = None,
        raise_exception: bool = True,
    ) -> SparkColumn:
        """
        Build Spark Column

        Parameters
        ----------
        df:
            Input DataFrame
        udfs:
            User-defined functions
        raise_exception
            If `True`, raise exception when input columns are not available,
            else, skip.
        """
        import pyspark.sql.functions as F
        from laktory.spark import functions as LF
        from laktory.spark.dataframe import has_column

        if udfs is None:
            udfs = []
        udfs = {f.__name__: f for f in udfs}

        # From SQL expression
        if self.sql_expression:
            logger.info(f"   {self.name}[{self.type}] as `{self.sql_expression}`)")
            col = F.expr(self.sql_expression).alias(self.name)
            if self.type not in ["_any"]:
                col = col.cast(self.type)
            return col

        # From Spark Function
        func_name = self.spark_func_name
        if func_name is None:
            func_name = "coalesce"

        # Get from UDFs
        f = udfs.get(func_name, None)
        if f is None:
            # Get from built-in spark functions
            f = getattr(F, func_name, None)

            if f is None:
                # Get from laktory functions
                f = getattr(LF, func_name, None)

        if f is None:
            raise ValueError(f"Function {func_name} is not available")

        _args = self.spark_func_args
        _kwargs = self.spark_func_kwargs

        logger.info(f"   {self.name}[{self.type}] as {func_name}({_args}, {_kwargs})")

        # Build args
        args = []
        for i, _arg in enumerate(_args):
            if df is not None:
                if _arg.is_column and not has_column(df, _arg.value):
                    logger.info(f"Column '{_arg.value}' not available")
                    if raise_exception:
                        raise ValueError(
                            f"Input column {_args} for {self.name} is not available"
                        )
                    else:
                        logger.info("Input columns not available. Skipping")
                        return None

            args += [_arg.to_spark()]

        # Build kwargs
        kwargs = {}
        for k, _arg in _kwargs.items():
            kwargs[k] = _arg.to_spark()

        # Function call
        col = f(*args, **kwargs)

        # Type Casting
        if self.type not in ["_any"]:
            col = col.cast(self.type)

        return col


if __name__ == "__main__":
    from laktory import models

    col = models.Column(
        name="open",
        spark_func_name="poly1",
        spark_func_args=[
            "data.open",
        ],
        spark_func_kwargs={"a": {"value": 2.0, "to_lit": True}},
        type="double",
    )

    spark_col = col.to_spark(df)



import os
import json

import pandas as pd
import pytest
import yaml

from laktory.models import Column
from laktory._testing import EventsManager

GOOGL = {
    "name": "close",
    "type": "double",
    "unit": "USD",
    "catalog_name": "dev",
    "schema_name": "markets",
    "table_name": "stock_prices",
}
data_dir = os.path.join(os.path.dirname(__file__), "data/")
if not os.path.exists(data_dir):
    os.makedirs(data_dir)
with open(os.path.join(data_dir, "googl.json"), "w") as fp:
    json.dump(GOOGL, fp)
with open(os.path.join(data_dir, "googl.yaml"), "w") as fp:
    yaml.dump(GOOGL, fp)

# Events
manager = EventsManager()
manager.build_events()
df = manager.to_spark_df()
df = df.filter("data.symbol == 'GOOGL'").limit(5)


def test_model():
    c0 = Column(**GOOGL)
    c1 = Column.model_validate(GOOGL)
    assert c1.type == "double"
    assert c1.catalog_name == "dev"
    assert c1.schema_name == "markets"
    assert c1.table_name == "stock_prices"
    assert c1.full_name == "dev.markets.stock_prices.close"
    assert "spark_func_name" in c1.model_fields
    assert c0 == c1


def test_read():
    c0 = Column(**GOOGL)

    with open(f"{data_dir}/googl.yaml", "r") as fp:
        c1 = Column.model_validate_yaml(fp)

    with open(f"{data_dir}/googl.json", "r") as fp:
        c2 = Column.model_validate_json_file(fp)

    assert c1 == c0
    assert c2 == c0


def test_spark():
    with pytest.raises(ValueError):
        close = Column(name="x", spark_func_args=("x",)).to_spark(df)

    c = Column(
        name="open",
        spark_func_args=[
            "data.open",
        ],
    )
    cs = c.to_spark(df)
    s = df.withColumn(c.name, cs).toPandas()[c.name].tolist()
    assert cs.__repr__() == "Column<'CAST(coalesce(data.open) AS STRING)'>"
    assert s == [
        "137.4600067138672",
        "135.44000244140625",
        "136.02000427246094",
        "133.58999633789062",
        "134.91000366210938",
    ]

    c = Column(
        name="open2",
        spark_func_name="poly1",
        spark_func_args=[
            "data.open",
        ],
        spark_func_kwargs={"a": {"value": 2.0, "to_lit": True}},
        type="double",
    )
    cs = c.to_spark(df)
    s = df.withColumn(c.name, cs).toPandas()[c.name].tolist()
    assert cs.__repr__() == "Column<'CAST(((2.0 * data.open) + 0.0) AS DOUBLE)'>"
    assert s == [
        274.9200134277344,
        270.8800048828125,
        272.0400085449219,
        267.17999267578125,
        269.82000732421875,
    ]

    c = Column(
        name="open_close",
        sql_expression="data.open*data.close",
        type="double",
    )
    cs = c.to_spark(df)
    s = df.withColumn(c.name, cs).toPandas()[c.name].tolist()
    assert (
        cs.__repr__()
        == "Column<'CAST((data.open * data.close) AS open_close AS DOUBLE)'>"
    )
    assert s == [
        18647.825014196802,
        18388.689710131846,
        18289.25068769534,
        18069.38217083132,
        18399.02695817873,
    ]

    c = Column(
        name="id",
        spark_func_name="coalesce",
        spark_func_args=[
            "data.`@id`",
        ],
        type="string",
    )
    cs = c.to_spark(df)
    s = df.withColumn(c.name, cs).toPandas()[c.name].tolist()
    assert cs.__repr__() == "Column<'CAST(coalesce(data.@id) AS STRING)'>"
    assert s == ["_id", "_id", "_id", "_id", "_id"]

    c = Column(
        name="no_explicit_type",
        spark_func_name="coalesce",
        spark_func_args=[
            "data.symbol",
        ],
        type="_any",
    )
    cs = c.to_spark(df)
    s = df.withColumn(c.name, cs).toPandas()[c.name].tolist()
    assert cs.__repr__() == "Column<'coalesce(data.symbol)'>"
    assert s == ["GOOGL", "GOOGL", "GOOGL", "GOOGL", "GOOGL"]

    c = Column(
        name="search_and_replace",
        spark_func_name="regexp_replace",
        spark_func_args=[
            "data.symbol",
            {"value": "O", "to_expr": False},
            {"value": "0", "to_expr": False},
        ],
        type="_any",
    )
    cs = c.to_spark(df)
    s = df.withColumn(c.name, cs).toPandas()[c.name].tolist()
    assert cs.__repr__() == "Column<'regexp_replace(data.symbol, O, 0, 1)'>"
    assert s == ["G00GL", "G00GL", "G00GL", "G00GL", "G00GL"]


def test_spark_udf():
    def x2(x):
        return 2 * x

    def x_square(x):
        return x * x

    c = Column(
        name="m2",
        type="double",
        spark_func_name="x2",
        spark_func_args=[
            "data.open",
        ],
    )
    cs = c.to_spark(df, udfs=[x2, x_square])
    s = df.withColumn(c.name, cs).toPandas()[c.name].tolist()
    assert cs.__repr__() == "Column<'CAST((data.open * 2) AS DOUBLE)'>"
    assert s == [
        274.9200134277344,
        270.8800048828125,
        272.0400085449219,
        267.17999267578125,
        269.82000732421875,
    ]


if __name__ == "__main__":
    test_model()
    test_read()
    test_spark()
    test_spark_udf()



import os
import logging
from datetime import datetime
from packaging import version

script_root = os.path.dirname(__file__)
package_name = "laktory"
laktory_root = f"./{package_name}/"


def main():
    # Filepaths
    version_filepath = os.path.join(laktory_root, "_version.py")
    local_env_filepath = os.path.join(script_root, "git.env")
    git_env_filepath = os.getenv("GITHUB_OUTPUT", local_env_filepath)
    changelog_filepath = os.path.join("./", "CHANGELOG.md")

    # Read version file
    with open(version_filepath) as fp:
        v0 = fp.read().split("=")[-1].strip().replace('"', "")
        v0 = version.parse(v0)
        v1 = version.Version(f"{v0.major}.{v0.minor}.{v0.micro + 1}")
    print(f"Bumping laktory to {v1}")

    # Update version file
    print(f"Updating _version.py")
    with open(version_filepath, "w") as fp:
        fp.write(f'VERSION = "{v1}"')

    # Set version as git action variable
    print(f"Setting git env var version {git_env_filepath}")
    with open(git_env_filepath, "a") as fp:
        fp.write(f"version={v1}")

    # Update CHANGELOG
    update_changelog(changelog_filepath, v0, v1)

    # Cleanup
    if git_env_filepath == local_env_filepath and os.path.exists(local_env_filepath):
        os.remove(local_env_filepath)


def update_changelog(changelog_filepath, v0, v1):
    # Update CHANGELOG
    print(f"Updating changelog")
    with open(changelog_filepath, "r") as fp:
        content = fp.read()

    # Set version line
    version_line = f"## [{v0}] - Unreleased"

    # Validate
    if version_line not in content:
        msg = f"CHANGELOG  does not include version {v0}"
        print("ValueError: " + msg)
        raise ValueError(msg)

    # Update current version release date
    today = datetime.utcnow()
    new_version_line = f"## [{v0}] - {today.strftime('%Y-%m-%d')}"
    content = content.replace(version_line, new_version_line)

    # Add placeholder for next version
    next_version_lines = [
        "# Release History",
        "",
        f"## [{v1}] - Unreleased",
        "### Added",
        "* n/a",
        "### Fixed",
        "* n/a",
        "### Updated",
        "* n/a",
        "### Breaking changes",
        "* n/a",
        "",
    ]
    content = content.replace("# Release History\n", "\n".join(next_version_lines))

    # Write
    with open(changelog_filepath, "w") as fp:
        fp.write(content)


if __name__ == "__main__":
    main()



from .stockprices import EventsManager
from .stockprices import StockPriceDataEvent
from .stockprices import StockPriceDataEventHeader
from .stockprices import StockPricesPipeline
from .stockprices import table_brz
from .stockprices import table_slv
from .stockprices import table_gld
from .stockprices import df_meta
from .stockprices import df_name
from .stockprices import table_slv_star
from .stockprices import view



try:
    import pyspark

    import laktory.spark.dataframe
    import laktory.spark.functions

    from pyspark.sql.dataframe import DataFrame
    from pyspark.sql.column import Column

except ModuleNotFoundError:
    # Mocks when pyspark is not installed

    class DataFrame:
        pass

    class Column:
        pass
